"""Define the grammar of a VBA source file and implement a syntactic parser."""

from typing import List, Union

from pyparsing import (Forward, ParseException, ParserElement, QuotedString,
                       Regex, Suppress, StringEnd, StringStart, Word,
                       delimitedList, infixNotation, nums, Or, Keyword,
                       FollowedBy)
from pyparsing import Optional as pOptional

from .abstract_syntax_tree import *
from .error import ParsingError
from .operator import OPERATORS_MAP
from .partial_block import *
from .preprocessor import Instruction, Preprocessor
from .type import Type, types

############
#  Tokens  #
############

# SEPARATOR

lparen = Suppress('(').setName('lparen')
rparen = Suppress(')').setName('rparen')
dot = Suppress('.').setName('dot')

# STATEMENT KEYWORD

call_kw = Suppress(Keyword('Call')).setName('call')
case_kw = Suppress(Keyword('Case')).setName('case')
close_kw = Suppress(Keyword('Close')).setName('close')
const_kw = Suppress(Keyword('Const')).setName('const')
declare_kw = Suppress(Keyword('Declare')).setName('declare')
defbool_kw = Suppress(Keyword('DefBool')).setName('defbool')
defbyte_kw = Suppress(Keyword('DefByte')).setName('defbyte')
defcur_kw = Suppress(Keyword('DefCur')).setName('defcur')
defdate_kw = Suppress(Keyword('DefDate')).setName('defdate')
defdbl_kw = Suppress(Keyword('DefDbl')).setName('defdbl')
defint_kw = Suppress(Keyword('DefInt')).setName('defint')
deflng_kw = Suppress(Keyword('DefLng')).setName('deflng')
deflnglng_kw = Suppress(Keyword('DefLngLng')).setName('deflnglng')
deflngptr_kw = Suppress(Keyword('DefLngPtr')).setName('deflngptr')
defobj_kw = Suppress(Keyword('DefObj')).setName('defobj')
defsng_kw = Suppress(Keyword('DefSng')).setName('defsng')
defstr_kw = Suppress(Keyword('DefStr')).setName('defstr')
defvar_kw = Suppress(Keyword('DefVar')).setName('defvar')
dim_kw = Suppress(Keyword('Dim')).setName('dim')
do_kw = Suppress(Keyword('Do')).setName('do')
else_kw = Suppress(Keyword('Else')).setName('else')
elseif_kw = Suppress(Keyword('ElseIf')).setName('elseif')
end_kw = Suppress(Keyword('End')).setName('end')
endif_kw = Suppress(Keyword('EndIf')).setName('endif')
enum_kw = Suppress(Keyword('Enum')).setName('enum')
erase_kw = Suppress(Keyword('Erase')).setName('erase')
event_kw = Suppress(Keyword('Event')).setName('event')
exit_kw = Suppress(Keyword('Exit')).setName('exit')
for_kw = Suppress(Keyword('For')).setName('for')
friend_kw = Suppress(Keyword('Friend')).setName('friend')
function_kw = Suppress(Keyword('Function')).setName('function')
get_kw = Suppress(Keyword('Get')).setName('get')
global_kw = Suppress(Keyword('Global')).setName('global')
gosub_kw = Suppress(Keyword('GoSub')).setName('gosub')
goto_kw = Suppress(Keyword('GoTo')).setName('goto')
if_kw = Suppress(Keyword('If')).setName('if')
implements_kw = Suppress(Keyword('Implements')).setName('implements')
input_kw = Suppress(Keyword('Input')).setName('input')
let_kw = Suppress(Keyword('Let')).setName('let')
lock_kw = Suppress(Keyword('Lock')).setName('lock')
loop_kw = Suppress(Keyword('Loop')).setName('loop')
lset_kw = Suppress(Keyword('LSet')).setName('lset')
next_kw = Suppress(Keyword('Next')).setName('next')
on_kw = Suppress(Keyword('On')).setName('on')
open_kw = Suppress(Keyword('Open')).setName('open')
option_kw = Suppress(Keyword('Option')).setName('option')
print_kw = Suppress(Keyword('Print')).setName('print')
private_kw = Suppress(Keyword('Private')).setName('private')
public_kw = Suppress(Keyword('Public')).setName('public')
put_kw = Suppress(Keyword('Put')).setName('put')
raiseevent_kw = Suppress(Keyword('RaiseEvent')).setName('raiseevent')
redim_kw = Suppress(Keyword('ReDim')).setName('redim')
resume_kw = Suppress(Keyword('Resume')).setName('resume')
return_kw = Suppress(Keyword('Return')).setName('return')
rset_kw = Suppress(Keyword('RSet')).setName('rset')
seek_kw = Suppress(Keyword('Seek')).setName('seek')
select_kw = Suppress(Keyword('Select')).setName('select')
set_kw = Suppress(Keyword('Set')).setName('set')
static_kw = Suppress(Keyword('Static')).setName('static')
stop_kw = Suppress(Keyword('Stop')).setName('stop')
sub_kw = Suppress(Keyword('Sub')).setName('sub')
type_kw = Suppress(Keyword('Type')).setName('type')
unlock_kw = Suppress(Keyword('Unlock')).setName('unlock')
wend_kw = Suppress(Keyword('Wend')).setName('wend')
while_kw = Suppress(Keyword('While')).setName('while')
with_kw = Suppress(Keyword('With')).setName('with')
write_kw = Suppress(Keyword('Write')).setName('write')

any_kw = Suppress(Keyword("Any")).setName('any')
as_kw = Suppress(Keyword("As")).setName('as')
byref_kw = Suppress(Keyword("ByRef")).setName('byref')
byval_kw = Suppress(Keyword("ByVal")).setName('byval')
case_kw = Suppress(Keyword("Case")).setName('case')
each_kw = Suppress(Keyword("Each")).setName('each')
error_kw = Suppress(Keyword("Error")).setName('error')
else_kw = Suppress(Keyword("Else")).setName('else')
in_kw = Suppress(Keyword("In")).setName('in')
new_kw = Suppress(Keyword("New")).setName('new')
shared_kw = Suppress(Keyword("Shared")).setName('shared')
until_kw = Suppress(Keyword("Until")).setName('until')
withevents_kw = Suppress(Keyword("WithEvents")).setName('withevents')
write_kw = Suppress(Keyword("Write")).setName('write')
optional_kw = Suppress(Keyword("Optional")).setName('optional')
paramarray_kw = Suppress(Keyword("ParamArray")).setName('paramarray')
preserve_kw = Suppress(Keyword("Preserve")).setName('preserve')
spc_kw = Suppress(Keyword("Spc")).setName('spc')
step_kw = Suppress(Keyword('Step')).setName('step')
tab_kw = Suppress(Keyword("Tab")).setName('tab')
then_kw = Suppress(Keyword("Then")).setName('then')
to_kw = Suppress(Keyword("To")).setName('to')

statement_keyword = call_kw | case_kw | close_kw | const_kw | declare_kw \
    | defbool_kw | defbyte_kw | defcur_kw | defdate_kw | defdbl_kw \
    | defint_kw | deflng_kw | deflnglng_kw | deflngptr_kw | defobj_kw \
    | defsng_kw | defstr_kw | defvar_kw | dim_kw | do_kw | else_kw \
    | elseif_kw | end_kw | endif_kw | enum_kw | erase_kw | event_kw \
    | exit_kw | for_kw | friend_kw | function_kw | get_kw | global_kw \
    | gosub_kw | goto_kw | if_kw | implements_kw | input_kw | let_kw \
    | lock_kw | loop_kw | lset_kw | next_kw | on_kw | open_kw | option_kw \
    | print_kw | private_kw | public_kw | put_kw | raiseevent_kw | redim_kw \
    | resume_kw | return_kw | rset_kw | seek_kw | select_kw | set_kw \
    | static_kw | stop_kw | sub_kw | type_kw | unlock_kw | wend_kw | while_kw \
    | with_kw | write_kw

marker_keyword = any_kw | as_kw | byref_kw | byval_kw | case_kw | each_kw \
    | else_kw | in_kw | new_kw | shared_kw | until_kw | withevents_kw \
    | write_kw | optional_kw | paramarray_kw | preserve_kw | spc_kw | step_kw \
    | tab_kw | then_kw | to_kw

# OPERATOR

addressof_kw = Keyword('AddressOf').setName('addressof')
and_kw = Keyword('And').setName('and')
eqv_kw = Keyword('Eqv').setName('eqv')
imp_kw = Keyword('Imp').setName('imp')
is_kw = Keyword('Is').setName('is')
like_kw = Keyword('Like').setName('like')
mod_kw = Keyword('Mod').setName('mod')
not_kw = Keyword('Not').setName('not')
or_kw = Keyword('Or').setName('or')
typeof_kw = Keyword('TypeOf').setName('typeof')
xor_kw = Keyword('Xor').setName('xor')

operator_keyword = addressof_kw | and_kw | eqv_kw | imp_kw | is_kw | like_kw \
    | mod_kw | not_kw | or_kw | typeof_kw | xor_kw

# TYPES

# Numbers
integer = Word(nums).setName("integer") \
    .setParseAction(lambda r: Literal(Type.Integer, r[0]))

# Boolean
true_kw = Keyword('True')
false_kw = Keyword('False')

boolean = (true_kw | false_kw).setName("boolean") \
    .setParseAction(lambda r: Literal(Type.Boolean, r[0]))

# String
string = QuotedString(quoteChar='"', escQuote='""').setName("string") \
    .setParseAction(lambda r: Literal(Type.String, r[0]))


literal = (integer | boolean | string).setName("literal")
literal_kw = true_kw | false_kw

# RESERVED

reserved = statement_keyword | marker_keyword | operator_keyword | literal_kw

# IDENTIFIER

identifier_regex = r"(?:[a-zA-Z]|_[a-zA-Z])[a-zA-Z0-9_]*"
identifier = (~reserved + Regex(identifier_regex)).setName("identifier") \
    .setParseAction(lambda r: Identifier(r[0]))
identifier_keyword = Regex(identifier_regex).setName("identifier keyword") \
    .setParseAction(lambda r: Identifier(r[0]))

# Types
variable_type = Or(types) | identifier

#############
#  Grammar  #
#############

expression = Forward().setName("expression")
statement = Forward().setName("statement")


#################
#  Expressions  #
#################

# Member access
def __build_recursive_member_access(expr, pos, result):
    tokens = list(result)
    node = tokens[0]

    for token in tokens[1:]:
        if isinstance(token, Identifier):
            node = Get(node, token)
        elif isinstance(token, FunCall):
            assert(isinstance(token.function, Identifier))
            node = FunCall(Get(node, token.function), token.arguments)

    return node


orphan_function_call_paren = Forward().setName("orphan_function_call_paren")
member_access_token = dot + (orphan_function_call_paren | identifier_keyword)
leading_member_access_token = member_access_token \
    + FollowedBy(member_access_token)
member_access = (
    (orphan_function_call_paren + FollowedBy(member_access_token) | identifier)
    + pOptional((leading_member_access_token)[...] + dot + identifier_keyword)
).setParseAction(__build_recursive_member_access)


# Operator
def __build_unary_operator(expr, pos, result):
    tokens = result[0].asList()
    assert(len(tokens) == 2)
    operator_symbol = tokens[0]
    operand = tokens[1]

    tree = UnOp(operator_symbol, operand)

    return tree


def __build_binary_operator(expr, pos, result):
    tokens = result[0].asList()
    operator_symbols = tokens[1::2]
    operands = tokens[0::2]

    tree = operands[0]
    for operator_symbol, operand in zip(operator_symbols, operands[1:]):
        tree = BinOp(operator_symbol, tree, operand)

    return tree


operators = OPERATORS_MAP.get_precedence_list(__build_unary_operator,
                                              __build_binary_operator)

# Function call
arguments_list_call = pOptional(delimitedList(expression)) \
    .setName("arguments_list_call") \
    .setParseAction(lambda r: ArgListCall(list(r)))
arguments_list_def = pOptional(delimitedList(identifier)) \
    .setName("arguments_list_def") \
    .setParseAction(lambda r: ArgListDef(list(r)))
orphan_function_call_paren << (identifier + lparen + arguments_list_call
                               + rparen) \
    .setName("orphan_function_call_paren") \
    .setParseAction(lambda r: FunCall(*r))
function_call_paren = (member_access + lparen + arguments_list_call + rparen) \
    .setName("function_call_paren") \
    .setParseAction(lambda r: FunCall(r[0], r[1]))
function_call_no_paren = (StringStart() + member_access + arguments_list_call
                          + StringEnd()) \
    .setName("function_call_no_paren") \
    .setParseAction(lambda r: FunCall(r[0], r[1]))

terminal = (literal | function_call_paren | member_access)
expression << infixNotation(
    terminal, operators, lpar=lparen, rpar=rparen) \
    .setName('expression')
expression_statement = (function_call_no_paren | expression) \
    .setName("expression_statement")

##################
#  Declarations  #
##################

# Variable
variable_declaration = (dim_kw + identifier
                        + pOptional(as_kw + variable_type
                                    + pOptional(Suppress('=') + expression))) \
    .setParseAction(lambda r: VarDec(*r)) \
    .setName("var dec")
variable_declaration_with_constructor = \
    (dim_kw + identifier + as_kw + new_kw + variable_type) \
    .setParseAction(lambda r: VarDec(*r, new=True))

# TODO difference between Set and Let
variable_assignment = (pOptional(let_kw | set_kw) + identifier
                       + Suppress('=') + expression) \
    .setParseAction(lambda r: VarAssign(*r)) \
    .setName("var assign")

# Function
procedure_scope = global_kw | public_kw | private_kw | friend_kw

procedure_header = (pOptional(procedure_scope) + pOptional(static_kw) + sub_kw
                    + identifier
                    + pOptional(lparen + arguments_list_def + rparen)
                    + pOptional(static_kw)) \
    .setParseAction(lambda r: ProcDefHeader(*r)) \
    .setName("proc header")
procedure_footer = (end_kw + sub_kw) \
    .setParseAction(lambda r: ProcDefFooter()) \
    .setName("proc footer")

function_header = (pOptional(procedure_scope) + pOptional(static_kw)
                   + function_kw + identifier
                   + pOptional(lparen + arguments_list_def + rparen)
                   + pOptional(as_kw + Suppress(variable_type))) \
    .setParseAction(lambda r: FunDefHeader(*r)) \
    .setName("function header")
function_footer = (end_kw + function_kw) \
    .setParseAction(lambda r: FunDefFooter()) \
    .setName("function footer")


declarative_statement = (
    variable_declaration_with_constructor | variable_declaration
    | variable_assignment
    | procedure_header | procedure_footer
    | function_header | function_footer) \
    .setName("declaration")

############################
#  Loops and conditionals  #
############################

# For
for_header = (for_kw + identifier + Suppress('=') + expression
              + to_kw + expression + pOptional(step_kw + expression)) \
    .setParseAction(lambda r: ForHeader(*r))
for_footer = (next_kw + pOptional(identifier)) \
    .setParseAction(lambda r: ForFooter(*r))

loop_statement = for_header | for_footer

# If
if_header = (if_kw + expression + pOptional(then_kw)) \
    .setParseAction(lambda r: IfHeader(*r))
elseif_header = (elseif_kw + expression + pOptional(then_kw)) \
    .setParseAction(lambda r: ElseIfHeader(*r))
else_header = else_kw.setParseAction(lambda r: ElseHeader())
if_footer = (end_kw + if_kw).setParseAction(lambda r: IfFooter())

if_oneliner = (if_kw + expression + then_kw + statement) \
    .setParseAction(lambda r: If(condition=r[0], body=[r[1]]))

conditional_statement = (if_oneliner | if_header | elseif_header | else_header
                         | if_footer)


####################
#  Error handling  #
####################

on_error = (on_kw + error_kw + ((goto_kw + (integer | identifier))
                                | (resume_kw + next_kw))) \
    .setParseAction(lambda r: OnError(r[0]) if len(r) > 0 else OnError())
resume = resume_kw + pOptional(next_kw | (integer | identifier)) \
    .setParseAction(lambda r: Resume(r[0] if len(r) > 0 else None))
error = (error_kw + integer).setParseAction(lambda r: ErrorStatement(r[0]))

error_statement = on_error | resume | error

# Wrap up
statement <<= declarative_statement | loop_statement | conditional_statement \
    | expression_statement | error_statement

############
#  Parser  #
############

# Packrat parsing cache parsing results, improving speed
ParserElement.enablePackrat(cache_size_limit=128)


class Parser:
    """
    Syntactic parser used to transform a list of instructions into an abstract
    syntax tree.
    """
    __nested_blocks: List[PartialBlock]

    def __init__(self) -> None:
        pass

    def build_ast(self, instructions: List[Instruction], file_name: str = "") \
            -> AST:
        """
        Main method of the Parser. Parses each of the instructions using a
        pyparsing grammar, and then process the results to handle multiline
        structures such as loops, conditionals, etc.
        """
        self.__nested_blocks = [PartialBlock()]
        for instruction in instructions:
            if instruction.instruction.strip() != '':
                parsed_instruction = self.__parse_instruction(instruction)
                if isinstance(parsed_instruction, (Statement, Block)):
                    self.__handle_statement(parsed_instruction)
                elif isinstance(parsed_instruction, BlockElement):
                    self.__handle_block_element(parsed_instruction)
                else:
                    assert (False)

        top_level_block = self.__nested_blocks.pop()
        main_sequence = Block(top_level_block.statements_blocks.pop())
        return main_sequence

    def __parse_instruction(self, instruction: Instruction)\
            -> Union[Statement, BlockElement]:
        """Use the pyparsing grammar to parse a single instruction."""
        try:
            parse_results = statement.parseString(
                instruction.instruction, parseAll=True)
        except ParseException as e:
            raise ParsingError(instruction.file_name, instruction.line_number,
                               str(e))

        return parse_results[0]

    def __handle_statement(self, statement: Statement) -> None:
        """
        Add a single-line statement to the last statements block of the
        innermost nest block.
        """
        current_block = self.__nested_blocks[-1]
        current_block.statements_blocks[-1].append(statement)

    def __handle_block_element(self, block_element: BlockElement) -> None:
        """
        Add a block element to the list of elements of the innermost partial
        block. Create it or add it to the statements of the parent block if
        needed.
        """
        if block_element.IsHeader:
            new_block = PartialBlock([block_element])
            self.__nested_blocks.append(new_block)
        else:
            self.__nested_blocks[-1].elements.append(block_element)
            if not block_element.IsFooter:
                self.__nested_blocks[-1].statements_blocks.append([])

        if block_element.IsFooter:
            complete_block = self.__nested_blocks.pop().build_block()
            parent_block = self.__nested_blocks[-1]
            parent_block.statements_blocks[-1].append(complete_block)

    @staticmethod
    def parse(content: str, file_name: Optional[str] = "") -> AST:
        """
        Parse the content of a file into an abstract syntax tree.

        :arg content: str content of the file
        :arg file_name: Optional name of the file
        :return: An AST representing the syntax of the file
        """
        parser = Parser()
        instructions = Preprocessor.preprocess(content, file_name)
        tree = parser.build_ast(instructions)

        return tree

    @staticmethod
    def parse_file(path: str) -> AST:
        """
        Parse a file into an abstract syntax tree.

        :arg path: Path of the file
        :return: An AST representing the syntax of the file
        """
        parser = Parser()
        instructions = Preprocessor.preprocess_file(path)
        tree = parser.build_ast(instructions)

        return tree



vbBack = "\x08"
vbCr = "\x0d"
vbCrLf = "\x0d\x0a"
vbFormFeed = "\x0c"
vbLf = "\x0a"
vbNewLine = "\n"
vbNullChar = "\x00"
vbTab = "\x09"
vbVerticalTab = "\x0b"
vbNullString = ""
vbObjectError = -2147221504



"""The memory is used to store the values of variables and functions."""

from dataclasses import dataclass
from enum import Enum
from itertools import count
from time import time
from typing import Any, Callable, Dict, List, Iterator

from .function import Function
from .value import Value
from .vba_class import Class


class Memory:
    """
    Represent the memory used for running a program, including static elements
    like functions and dynamic ones like local variables.
    """
    global_variables: Dict[str, Value]
    _local_variables: List[Dict[str, Value]]
    _functions: Dict[str, Function]
    _classes: Dict[str, Class]

    def __init__(self) -> None:
        self.global_variables = dict()
        self._local_variables = []
        self._functions = dict()
        self._classes = dict()

    def add_function(self, full_name: str, function: Function) -> None:
        self._functions[full_name] = function

    def get_function(self, full_name: str) -> Function:
        return self._functions[full_name]

    def set_variable(self, local_name: str, value: Value) -> None:
        self._local_variables[-1][local_name] = value

    def get_variable(self, name: str) -> Value:
        # TODO use full name for variable storage
        # TODO use variable reference instead of name to ensure unicity
        local_name = name.split('.')[-1]
        try:
            return self.locals[local_name]
        except KeyError:
            return self.global_variables[name]

    @property
    def locals(self):
        return self._local_variables[-1]

    @property
    def functions(self):
        return self._functions

    @property
    def classes(self):
        return self._classes

    def new_locals(self) -> None:
        self._local_variables.append(dict())

    def discard_locals(self) -> None:
        self._local_variables.pop()


# TODO test hooks
class OutsideWorld:
    """
    Represent the interaction between a running program and an emulated
    exterior environment.
    """
    @dataclass
    class Event:
        """
        Class used to record usefull events. It stores its category, the
        occuring time and a corresponding identifier, its context (the name of
        the symbol that triggered the event), and a data field, which depends
        on the event category.
        """
        class Category(Enum):
            STDOUT = "stdout"
            FILESYSTEM = "filesystem"
            NETWORK = "network"
            EXECUTION = "execution"
            ENVIRONMENT = "environment"

        identifier: int
        time: float
        category: "OutsideWorld.Event.Category"
        context: str
        data: Any

        def to_dict(self, reproducible: bool = False) -> Dict[str, Any]:
            """
            Return the dict of the event attributes.

            :arg reproducible: If True, don't add the time field
            :returns: The event attributes
            """
            d = {
                "identifier": self.identifier,
                "category": self.category.value,
                "context": self.context,
                "data": self.data,
            }

            if not reproducible:
                d["time"] = self.time

            return d

    __start_time: float
    __events_count: Iterator[int]
    __hooks: Dict["OutsideWorld.Event.Category",
                  Callable[["OutsideWorld.Event"], None]]
    events: List["OutsideWorld.Event"]
    files: Dict[str, str]

    def __init__(self) -> None:
        self.__start_time = time()
        self.__events_count = count()
        self.__hooks = {event_category: OutsideWorld.nop
                        for event_category in OutsideWorld.Event.Category}
        self.events = []
        self.files = dict()

    def add_event(self, category: "OutsideWorld.Event.Category", context: str,
                  data: Any) -> None:
        """
        Record a new event, calling the corresponding hook.
        """
        t = time() - self.__start_time
        identifier = next(self.__events_count)
        event = OutsideWorld.Event(identifier, t, category, context, data)

        self.events.append(event)
        self.__hooks[category](event)

        if category is OutsideWorld.Event.Category.FILESYSTEM:
            if data['type'] == 'Write':
                path = data['path']
                content = self.files.get(path, "")
                self.files[path] = content + data['data']

    def to_dict(self, reproducible: bool = False) -> Dict[str, Any]:
        """
        Return the list of events and files in a dict.

        :arg reproducible: If True, discard the time field of the events
        :returns: A dict with keys 'events' and 'files'
        """
        d: Dict[str, Any] = dict()
        d['events'] = [event.to_dict(reproducible)
                       for event in self.events]
        d['files'] = self.files

        return d

    def add_hook(self, category: "OutsideWorld.Event.Category",
                 hook: Callable[["OutsideWorld.Event"], None]) -> None:
        """
        Add a hook that will be called for each recording of a given event
        category.

        Hooks should not be lambda's for OutsideWorld to be used with
        Serializer.
        """
        self.__hooks[category] = hook

    @staticmethod
    def nop(event: "OutsideWorld.Event") -> None:
        """Function that can be used as a hook that does nothing."""
        pass



"""
Provide abstract base classes to implement the Visitor design pattern.

A Visitable object has nothing more to do than inheriting Visitable. All the
algorithm must be only defined in the Visitor object. To handle a special
Visitable type SubVisitable, a Visitor must implement visit_SubVisitable, and
can then call visit(SubVisitable).
"""


from abc import abstractmethod


class Visitable:
    """Base class for Visitable classes, must simply be inherited."""
    @abstractmethod
    def __init__(self):
        pass

    def accept(self, visitor: "Visitor") -> None:
        assert(isinstance(visitor, Visitor))

        visiter = "visit_" + type(self).__qualname__.replace(".", "_")
        try:
            visiter_function = getattr(visitor, visiter)
        except AttributeError:
            msg = f"Visitor {type(visitor).__qualname__} doesn't handle " \
                + f"Visitable type {type(self).__qualname__}"
            raise NotImplementedError(msg)

        return visiter_function(self)


class Visitor:
    """
    Base class for Visitor classes. For a subclass to be able to handle the
    Visitable type T, it must implement a method
    visit_T(self, visitable : Visitable) -> None. Then, to apply the algorithm
    to a Visitable object visitable, simply call visitor.visit(visitable).
    """
    @abstractmethod
    def __init__(self):
        pass

    def visit(self, visitable: Visitable) -> None:
        """
        Apply an algorithm to a Visitable object of type T. The algorithm must
        be implemented in visit_T, otherwise a NotImplementedError is thrown.
        """
        assert(isinstance(visitable, Visitable))

        visitable.accept(self)



"""Pseudo-compilation stage: allow to extract references and function body."""

from dataclasses import dataclass
from importlib.util import spec_from_file_location, module_from_spec
from inspect import getmembers, isclass, isfunction
from itertools import chain
from pathlib import Path
from types import ModuleType
from typing import Dict, Any, List

from .reference import *
from .abstract_syntax_tree import *
from .error import ResolutionError
from .function import ExternalFunction, InternalFunction
from .side_effect import Memory
from .syntax import Parser
from .vba_class import Class
from .visitor import Visitor
from .value import Value


@dataclass
class Program:
    """
    Represent a statically analysed program : a set of references, with a
    memory containing already initialized values.
    """
    memory: Memory
    environment: Environment

    def to_dict(self) -> Dict[str, Any]:
        d = dict()
        d['memory'] = {'functions': list(self.memory._functions.keys()),
                       'classes': list(self.memory._classes.keys())}
        d['environment'] = self.environment.to_dict()
        return d


@dataclass
class Unit:
    """
    Compilation unit, representing a single file, with its content and type.
    """
    Type = Enum("Type", "Class Procedures Script")

    content: str
    unit_type: 'Unit.Type'
    name: Optional[str] = ""

    @staticmethod
    def from_file(file_path: str) -> 'Unit':
        """
        Build a unit from a file path. See Unit.from_content for information
        about extensions.
        """
        with open(file_path) as f:
            content = f.read()

        return Unit.from_content(content, file_path)

    @staticmethod
    def from_content(file_content: str, file_path: Optional[str] = None) \
            -> 'Unit':
        """
        Build a unit from a file content. The unit type is determined based on
        the extension of the file: 'cls' for class module, 'bas' for
        procedures, 'vbs' for script.
        """
        path = Path(file_path)
        extension = path.suffix[1:]
        if extension == "cls":
            unit_type = Unit.Type.Class
        elif extension == "bas":
            unit_type = Unit.Type.Procedures
        elif extension == "vbs":
            unit_type = Unit.Type.Script
        else:
            msg = f"File {path} has unrecognized extension {extension}"
            raise CompilationError(msg)

        return Unit(file_content, unit_type, path.stem)


class Compiler(Visitor):
    """
    Class used for references extraction. You can analyse several modules in a
    row by using add_module multiple times, and can then retrieve the built
    references and memory by accessing the corresponding properties. To start a
    new analysis, use reset.
    """
    __environment: Environment
    __current_reference: Reference
    __memory: Memory

    def __init__(self) -> None:
        self.reset()

    def reset(self) -> None:
        """Reset the state of the compiler, erasing references and memory."""
        self.__environment = Environment()
        self.__current_reference = self.__environment
        self.__memory = Memory()

    def add_project(self, project_name: str) -> None:
        """
        Add a project to the current program and use it as the parent project
        for the following operations.
        """
        project = self.__environment.build_child(Project,
                                                 name=project_name)
        self.__current_reference = project

    def add_module(self, ast: AST, module_type,
                   module_name: str = None) -> None:
        """
        Add a module to the current project, analysing the AST of the module.
        """
        if isinstance(self.__current_reference, Module):
            self.__current_reference = self.__current_reference.parent
        elif type(self.__current_reference) is Project:
            pass
        elif type(self.__current_reference) is Environment:
            try:
                project = self.__environment.get_child("Default")
            except ResolutionError:
                project = self.__environment.build_child(
                    Project,
                    name="Default")

            self.__current_reference = project

        module = self.__current_reference.build_child(
            module_type, name=module_name)

        if module_type is ClassModule:
            self.__memory.classes[str(module)] = Class([], module)

        self.__current_reference = module
        self.visit(ast)

    def load_host_project(self, project_path: str) -> None:
        """Load a Python package as a VBA host project."""
        path = Path(project_path)
        assert(path.is_dir())
        project_name = path.name

        project = Project(project_name)
        self.__environment.add_child(project)
        self.__current_reference = project

        for module_path in path.glob("*.py"):
            module_name = f"{project_name}.{module_path.name[:-3]}"
            module_spec = spec_from_file_location(module_name,
                                                  module_path.absolute())
            module = module_from_spec(module_spec)
            module_spec.loader.exec_module(module)

            self.load_host_module(module)

        self.__current_reference = self.__current_reference.parent

    def load_host_module(self, module: ModuleType) -> None:
        """
        Extract objects defined in an already loaded Python module and add them
        to the current program.
        """
        def locally_defined(predicate):
            def decorated_predicate(member):
                return predicate(member) \
                    and member.__module__ == module.__name__

            return decorated_predicate

        module_name = module.__name__.split('.')[-1]

        classes = getmembers(module, locally_defined(isclass))
        assert(len(classes) in (0, 1))
        if len(classes) == 0:
            self.__current_reference = self.__current_reference.build_child(
                ProceduralModule, module_name)
            functions_holder = module
        else:
            self.__current_reference = self.__current_reference.build_child(
                ClassModule, module_name)
            class_name = classes[0][0]
            py_class = getattr(module, class_name)
            functions_holder = py_class

            for variable in py_class.variables:
                self.__current_reference.build_child(
                    Variable,
                    variable,
                    extent=Variable.Extent.Object
                )

            vba_class = Class(py_class.variables, self.__current_reference)
            self.__memory.classes[str(self.__current_reference)] = vba_class

        for name, function in getmembers(
                functions_holder, locally_defined(isfunction)):
            self.load_host_function(function)

        for name, python_value in getmembers(
                module, lambda obj: not(isfunction(obj) or isclass(obj))):
            if not name.startswith('__'):
                reference = self.__try_add_variable(name)
                vba_value = Value.from_python_base_type(python_value)
                self.__memory.global_variables[str(reference)] = vba_value

        self.__current_reference = self.__current_reference.parent

    def load_host_function(self, function: ExternalFunction.Signature) \
            -> None:
        typed_function = ExternalFunction.from_function(function)
        name = typed_function.name

        function = self.__current_reference.build_child(
            FunctionReference, name=name)
        typed_function.reference = function
        self.__memory.add_function(str(function), typed_function)

    @property
    def program(self):
        """Return the compiled program."""
        return Program(self.__memory, self.__environment)

    def __parse_callable(self, definition: Union[FunDef, ProcDef]) -> None:
        # Extract information
        name = definition.name.name
        if definition.arguments is None:
            args = []
        else:
            args = list(map(lambda t: t.name, definition.arguments.args))
        body = definition.body

        # Build reference and memory representation
        function_ref = FunctionReference(name)
        self.__current_reference.add_child(function_ref)
        function_object = InternalFunction(name, args, body, function_ref)

        # Add them
        self.__memory.add_function(str(function_ref), function_object)

        self.__current_reference = function_ref

        for arg in args:
            self.__try_add_variable(arg)

        if type(definition) is FunDef:
            self.__try_add_variable(name)

        self.__try_add_variable(name)
        self.visit_Block(definition)
        self.__current_reference = self.__current_reference.parent

    def __try_add_variable(self, name: str) -> None:
        assert(isinstance(self.__current_reference,
                          (Module, FunctionReference)))

        try:
            self.__current_reference.get_child(name)
            return
        except ResolutionError:
            pass

        if isinstance(self.__current_reference, ProceduralModule):
            extent = Variable.Extent.Module
        elif isinstance(self.__current_reference, ClassModule):
            extent = Variable.Extent.Object
            vba_class = self.__memory.classes[str(self.__current_reference)]
            vba_class.variables.append(name)
        elif isinstance(self.__current_reference, FunctionReference):
            extent = Variable.Extent.Procedure

        return self.__current_reference.build_child(
            Variable,
            name=name,
            extent=extent)

    def visit_Block(self, block: Block) -> None:
        for statement in block.body:
            self.visit(statement)

    def visit_VarDec(self, var_dec: VarDec) -> None:
        self.__try_add_variable(var_dec.identifier.name)

    def visit_VarAssign(self, var_assign: VarAssign) -> None:
        if type(var_assign.variable) is Identifier:
            name = var_assign.variable.name
            self.__try_add_variable(name)

    def visit_FunDef(self, fun_def: FunDef) -> None:
        self.__parse_callable(fun_def)

    def visit_ProcDef(self, proc_def: ProcDef) -> None:
        self.__parse_callable(proc_def)

    def visit_FunCall(self, fun_call: FunCall) -> None:
        pass

    def visit_If(self, if_block: If) -> None:
        self.visit_Block(if_block)
        for elseif in if_block.elsifs:
            self.visit_Block(elseif)
        if if_block.else_block is not None:
            self.visit_Block(if_block.else_block)

    def visit_For(self, for_loop: For) -> None:
        self.__try_add_variable(for_loop.counter.name)
        self.visit_Block(for_loop)

    def visit_OnError(self, on_error: OnError) -> None:
        pass

    def visit_Resume(self, resume: Resume) -> None:
        pass

    def visit_ErrorStatement(self, error: ErrorStatement) -> None:
        pass

    @staticmethod
    def compile_unit(unit: Unit) -> Program:
        return Compiler.compile_units([unit])

    @staticmethod
    def compile_units(
            units: List[Unit], project: Optional[str] = None) -> Program:
        """
        Compile a list of units belonging to the same project.
        """
        compiler = Compiler()

        if project is not None:
            compiler.add_project(project)

        for unit in units:
            if unit.unit_type is Unit.Type.Class:
                module_type = ClassModule
            else:
                module_type = ProceduralModule

            ast = Parser.parse(unit.content, unit.name)
            compiler.add_module(ast, module_type, unit.name)

        return compiler.program

    @staticmethod
    def compile_file(path: str) -> Program:
        return Compiler.compile_files([path])

    @staticmethod
    def compile_files(
            paths: List[str], project: Optional[str] = None) -> Program:
        """
        Parse and compile a list of files with Office extensions (cls and bas)
        or vbs extension for standalone scripts.
        """
        units = [Unit.from_file(path) for path in paths]
        return Compiler.compile_units(units, project)

    @staticmethod
    def compile_project(project_path: str) -> Program:
        """
        Recursively compile a directory, only taking into accound cls and bas
        files.
        """
        path = Path(project_path)
        assert(path.is_dir())
        paths = list(chain(list(path.rglob(f'*.{ext}'))
                           for ext in ('cls', 'bas')))
        paths = [str(file.absolute()) for glob in
                 (path.rglob(f"*.{ext}") for ext in ('cls', 'bas'))
                 for file in glob]

        return Compiler.compile_files(paths, path.stem)

    @staticmethod
    def link_libraries(program: Program, libraries: List[str]) -> Program:
        """
        Link host libraries to an already compiled program.

        :arg program: Program to link the library to
        :arg library: Paths of the libraries
        :returns: The linked program
        """
        compiler = Compiler()
        compiler.__memory = program.memory
        compiler.__environment = program.environment
        compiler.__current_reference = program.environment

        for library in libraries:
            compiler.load_host_project(library)

        return compiler.program



"""Define parts of AST nodes, useful for block statements."""

from abc import abstractmethod, ABC
from typing import List, Union

from .abstract_syntax_tree import (For, ElseIf, If, Block, ProcDef, FunDef,
                                   Statement)


class BlockElement(ABC):
    """
    Structural element of a block, for example an ElseIf line in an If block,
    only used during the parsing process.
    """
    IsHeader: bool = False
    IsFooter: bool = False

    @abstractmethod
    def __init__(self) -> None:
        pass


class ForHeader(BlockElement):
    IsHeader = True

    def __init__(self, counter, start, end, step=None):
        super().__init__()
        self.counter = counter
        self.start = start
        self.end = end
        self.step = step


class ForFooter(BlockElement):
    IsFooter = True

    def __init__(self, counter=None):
        super().__init__()
        self.counter = counter


class IfHeader(BlockElement):
    IsHeader = True

    def __init__(self, condition):
        super().__init__()
        self.condition = condition


class ElseIfHeader(BlockElement):
    def __init__(self, condition):
        super().__init__()
        self.condition = condition


class ElseHeader(BlockElement):
    def __init__(self):
        super().__init__()


class IfFooter(BlockElement):
    IsFooter = True

    def __init__(self):
        super().__init__()


class ProcDefHeader(BlockElement):
    IsHeader = True

    def __init__(self, name, arguments=None):
        super().__init__()
        self.name = name
        self.arguments = arguments


class ProcDefFooter(BlockElement):
    IsFooter = True

    def __init__(self):
        super().__init__()


class FunDefHeader(BlockElement):
    IsHeader = True

    def __init__(self, name, arguments=None):
        super().__init__()
        self.name = name
        self.arguments = arguments


class FunDefFooter(BlockElement):
    IsFooter = True

    def __init__(self):
        super().__init__()


class PartialBlock:
    """
    A PartialBlock is used in the process of building block statements,
    e.g. function definitions or conditionals. It handles statements with
    multiple sub-blocks, as in a if/elseif/else structure.
    """
    elements: List[BlockElement]
    statements_blocks: List[List[Union[Statement, Block]]]

    def __init__(self, elements: List[BlockElement] = None,
                 statements_blocks: List[List[Statement]] = None) -> None:
        super().__init__()
        self.elements = elements if elements is not None else []
        self.statements_blocks = statements_blocks \
            if statements_blocks is not None \
            else [[]]  # type: ignore[assignment]

    # TODO add syntax check
    def build_block(self) -> Block:
        """
        Build the Block obtained after pulling together all the BlockElement
        and Statements it is made of.
        """
        header = self.elements[0]
        if isinstance(header, ForHeader) and len(self.elements) == 2:
            for_block = For(header.counter, header.start, header.end,
                            header.step, body=self.statements_blocks[0])
            return for_block

        elif isinstance(header, IfHeader):
            blocks = zip(self.statements_blocks, self.elements)
            if_body, if_header = next(blocks)
            assert (isinstance(if_header, IfHeader))
            if_block = If(condition=if_header.condition, body=if_body)

            for statements, element in blocks:
                if isinstance(element, ElseIfHeader):
                    elseif = ElseIf(
                        condition=element.condition, body=statements)
                    if_block.elsifs.append(elseif)
                elif isinstance(element, ElseHeader):
                    else_block = Block(statements)
                    if_block.else_block = else_block
                elif isinstance(element, IfFooter):
                    break

            return if_block

        elif isinstance(header, ProcDefHeader):
            body = self.statements_blocks.pop()
            proc_def = ProcDef(header.name, header.arguments, body=body)
            return proc_def

        elif isinstance(header, FunDefHeader):
            body = self.statements_blocks.pop()
            fun_def = FunDef(header.name, header.arguments, body=body)
            return fun_def

        else:
            raise NotImplementedError()



"""Define the different VBA types."""

# To implement a new type, you must :
#  - add it to the list of Type fields
#  - add a matching grammar rule in syntax.py if needed
#  - update value.py to allow evaluation
#
# TODO:
#  - arrays
#  - Byte
#  - Char
#  - Date
#  - Decimal
#  - Double
#  - Long
#  - Object
#  - SByte
#  - Short
#  - Single
#  - String
#  - UInteger
#  - ULong
#  - User-Defined
#  - UShort

from enum import Enum

Type = Enum("Type", "Integer Boolean String Object Function")
types = tuple(Type.__members__.keys())



"""
Preprocessing of the VBS file : handling of multiple instructions per line,
comments, etc.
"""

# Specifications reference:
# https://docs.microsoft.com/en-us/dotnet/visual-basic/programming-guide/language-features/statements
#
# Multiple instructions/lines:
#  - Separation with colon DONE
#  - Line ending in underscore DONE
#  - Implicit continuation TODO:
#       - Comma, concatenation operator DONE
#       - Parenthesis, brackets TODO
#       - Embedded expression TODO
#       - Assignment operators TODO
#       - Binary operators within an expression TODO
#       - Is and IsNot operators TODO
#       - Member qualifier character TODO
#       - XML property qualifier TODO
#       - < and > when specifying an attribute TODO
#       - before and after query operations TODO
#       - after IN in a For Each TODO
#       - after From in a collection initializer TODO
#
# Comments:
#  - Comments added in Instruction pretty-print TODO

from typing import Any, Dict, List, Optional

from .error import PreprocessorError


class Instruction:
    """A single instruction, stored with its context."""
    instruction: str
    multiline: str
    single: bool
    file_name: str
    line_number: int

    def __init__(self, instruction: str, multiline: str, single: bool,
                 file_name: str, line_number: int) -> None:
        """
        :arg instruction: Unique instruction
        :arg multiline: Group of continuing lines the instruction is part of
        :arg single: Specify if the instruction is the only one in the
        multiline
        :arg file_name: Name of the file
        :arg line_number: Line of the instruction in the file
        """
        self.instruction = instruction
        self.multiline = multiline
        self.single = single
        self.file_name = file_name
        self.line_number = line_number

    def __str__(self) -> str:
        indent = "    "
        s = f"{self.file_name}:{self.line_number}:{indent}{self.instruction}"
        if not self.single:
            s += indent + "in\n" + indent
            s += self.multiline.replace('\n', '\n' + indent)

        return s

    def __repr__(self) -> str:
        return f"Instruction({self.instruction})"

    def to_dict(self) -> Dict[str, Any]:
        return self.__dict__


class Preprocessor:
    """
    Used to extract single instructions from the content of a source file.
    """
    CONTINUING_LINE_DELIMITER = " _"
    CONTINUING_LINE_OPERATORS = (",", "&")
    COMMENT_DELIMITERS = ("REM", "'")
    MULTIPLE_INSTRUCTIONS_DELIMITER = ":"

    def __init__(self) -> None:
        pass

    def extract_instructions(self, file_name: str, file_content: str) \
            -> List[Instruction]:
        """
        Extract all the individual instructions from a file.

        :arg file_name: Name of the file
        :arg file_content: Content of the file
        :returns: The list of instructions, in the same order as in the file
        """
        self.__instructions: List[Instruction] = []
        self.__line_number = 1
        self.__instruction_line = 1
        self.__continuing_line = False
        self.__multiline = ""
        self.__concatenated_line = ""
        self.__file_name = file_name

        for line in file_content.splitlines():
            self.__handle_line(line)

        return self.__instructions

    def __handle_line(self, line: str) -> None:
        # Handle comments
        commentless_line = self.__remove_comments(line).strip()

        # Concatenate multi-line instructions
        if self.__continuing_line:
            self.__multiline += "\n"
        else:
            self.__concatenated_line = ""
            self.__instruction_line = self.__line_number
        self.__concatenated_line += commentless_line
        self.__multiline += line

        # If the instruction has not come to its end yet
        if commentless_line.endswith(self.CONTINUING_LINE_DELIMITER):
            self.__continuing_line = True
            del_chars = len(self.CONTINUING_LINE_DELIMITER)
            self.__concatenated_line = self.__concatenated_line[:-del_chars]
        elif any(commentless_line.strip().endswith(op)
                 for op in self.CONTINUING_LINE_OPERATORS):
            self.__continuing_line = True
            # If the instruction is complete
        else:
            instructions = self.__split_line(self.__concatenated_line)
            single = len(instructions) == 1 \
                and "\n" not in self.__multiline
            for instruction in instructions:
                self.__add_instruction(instruction, single)

            self.__continuing_line = False
            self.__concatenated_line = ""
            self.__multiline = ""

        self.__line_number += 1

    def __add_instruction(self, instruction: str, single: bool) -> None:
        self.__instructions.append(Instruction(
            instruction=instruction.strip(),
            multiline=self.__multiline,
            single=single,
            file_name=self.__file_name,
            line_number=self.__instruction_line
        ))

    @classmethod
    def __remove_comments(cls, line: str) -> str:
        """
        Search for the first REM or ' character not enclosed in a string or in
        another token
        """
        in_string = False
        position = 0

        if any(line.startswith(f"{d} ") for d in cls.COMMENT_DELIMITERS):
            return ''

        while position < len(line):
            char = line[position]
            remaining_line = line[position:]

            if in_string:
                if remaining_line.startswith('""'):
                    position += 1
                elif char == '"':
                    in_string = False
            else:
                if any(remaining_line.startswith(f" {d} ")
                       for d in cls.COMMENT_DELIMITERS):
                    return line[:position]
                elif char == '"':
                    in_string = True

            position += 1

        return line

    @classmethod
    def __split_line(cls, line: str) -> List[str]:
        """
        Split a line containing no comments into its different instructions,
        ignoring separators inside strings.
        """
        in_string = False
        instruction_start = 0
        position = 0
        delim = cls.MULTIPLE_INSTRUCTIONS_DELIMITER
        instructions = []

        while position < len(line):
            char = line[position]
            remaining_line = line[position:]

            if in_string:
                if remaining_line.startswith('""'):
                    position += 1
                elif char == '"':
                    in_string = False
            else:
                if remaining_line.startswith(delim):
                    instructions.append(line[instruction_start:position])
                    instruction_start = position + 1
                elif char == '"':
                    in_string = True

            position += 1

        instructions.append(line[instruction_start:position])

        return instructions

    @staticmethod
    def preprocess(file_content: str, file_name: Optional[str] = None) \
            -> List[Instruction]:
        """
        Extract the executable instructions of a file, returning them as a
        list.
        """
        preprocessor = Preprocessor()
        return preprocessor.extract_instructions(file_name, file_content)

    @staticmethod
    def preprocess_file(file_path: str) -> List[Instruction]:
        with open(file_path, 'r') as f:
            file_content = f.read()

        return Preprocessor.preprocess(file_content, file_path)



"""Defines the logic of built-in VBA operators"""

# To implement support for a new operator, you must:
#  - ensure the types it works with are implemented, see type.py
#  - add a new entry to OPERATORS_MAP
#  - add it to the list of operator tokens

# TODO:
#  - Take Let-coercion into account, improve conversion process
#  - Extends supported types
#  - Implement Like and Is

from abc import ABC, abstractmethod
from dataclasses import InitVar, dataclass
from inspect import signature
from typing import (Any, Callable, Dict, Generic, List, Optional, Tuple,
                    TypeVar, Union)
from typing import Type as pType

from pyparsing import Or, opAssoc, ParseExpression

from .error import OperatorError
from .type import Type
from .value import ConversionError, Value


class Operation(ABC):
    pass


@dataclass
class UnaryOperation(Operation):
    """Represent the underlying function of a unary operator"""
    function: InitVar[Callable[[Any], Any]]
    arg_type: Type
    return_type: Type

    # Workaround for mypy not supporting class-level hints for callbacks
    def __post_init__(self, function: Callable[[Any], Any]) -> None:
        self.function = function


@dataclass
class BinaryOperation(Operation):
    """Represent the underlying function of a binary operator"""
    function: InitVar[Callable[[Any, Any], Any]]
    left_type: Type
    right_type: Type
    return_type: Type

    # Workaround for mypy not supporting class-level hints for callbacks
    def __post_init__(self, function: Callable[[Any, Any], Any]) -> None:
        self.function = function


T = TypeVar('T', UnaryOperation, BinaryOperation)


class Operator(Generic[T], ABC):
    """
    Abstract operator, be it unary or binary.
    """
    symbol: str
    operations: List[T]

    @abstractmethod
    def __init__(self, symbol: str, operations: List[T]) -> None:
        self.symbol = symbol
        self.operations = operations

    @staticmethod
    def from_symbol(symbol: str) -> "Operator":
        """Build an operator from its symbol. It uses OPERATOR_MAP."""
        try:
            return OPERATORS_MAP[symbol]
        except KeyError:
            msg = f"Operator {symbol} is not supported yet"
            raise NotImplementedError(msg)

    def __str__(self) -> str:
        return self.symbol

    @staticmethod
    def from_operation(symbol: str, operation: Operation) -> "Operator":
        """Build an operator from a symbol and a single."""
        operator_type: Union[pType[BinaryOperator], pType[UnaryOperator]]
        if isinstance(operation, UnaryOperation):
            return UnaryOperator(symbol, [operation])
        elif isinstance(operation, BinaryOperation):
            return BinaryOperator(symbol, [operation])

        msg = f"Can't build Operator from type {type(operation)}"
        raise RuntimeError(msg)


class UnaryOperator(Operator[UnaryOperation]):
    """
    Class used to evaluate unary operations with operators that may accept
    different types.
    """

    def __init__(self, symbol: str, operations: List[UnaryOperation]) -> None:
        super().__init__(symbol, operations)

    def operate(self, value: Value) -> Value:
        """
        Find the first UnaryOperation in self.operations that can be called
        with the given value, and return its result. Raise a ConversionError if
        there is no compatible operator.
        """
        for operation in self.operations:
            try:
                converted_arg = value.convert_to(operation.arg_type)
                value = operation.function(converted_arg)
                value_type = operation.return_type
                return Value.from_value(value, value_type)
            except ConversionError:
                pass

        msg = f"Type {value.base_type} does not match with operator " \
              f"{self.symbol}"
        raise OperatorError(msg)


class BinaryOperator(Operator[BinaryOperation]):
    """
    Class used to evaluate binary operations with operators that may accept
    different types.
    """

    def __init__(self, symbol: str, operations: List[BinaryOperation]) -> None:
        super().__init__(symbol, operations)

    def operate(self, left_value: Value, right_value: Value) -> Value:
        """
        Find the first BinaryOperation in self.operations that can be called
        with the two given values, and return its result. Raise a
        ConversionError if there is no compatible operator.
        """
        for operation in self.operations:
            try:
                converted_left = left_value.convert_to(operation.left_type)
                converted_right = right_value.convert_to(operation.right_type)
                value = operation.function(converted_left, converted_right)
                value_type = operation.return_type
                return Value.from_value(value, value_type)
            except ConversionError:
                pass

        msg = f"Types {left_value.base_type}, {right_value.base_type} do " \
              f"not match with operator {self.symbol}"
        raise OperatorError(msg)


class OperatorsMap:
    """
    Main class used to add operators to the parser. A single global instance is
    used, OPERATORS_MAP.

    To implement support for a new operator, use add_operation or the <<
    operator. Precedence is managed by start_precedence_group and
    end_precedence_group. Finally, you can use get_precedence_list with
    pyparsing.infixNotation.
    """
    __operators: Dict[str, Operator]
    __ordered_operators: List[List[str]]
    __growing_precedence: bool

    def __init__(self) -> None:
        self.__operators = dict()
        self.__ordered_operators = []
        self.__growing_precedence = True

    def start_precedence_group(self) -> None:
        """
        Start a precedence group: all the following added operations will have
        the same precedence.
        """
        self.__ordered_operators.append([])
        self.__growing_precedence = False

    def end_precedence_group(self) -> None:
        """
        Stop the current precedence group: all the added operations will have
        increasing precedence.
        """
        self.__growing_precedence = True

    def get_precedence_list(self, parse_unary, parse_binary):
        """Build a precedence list to be used with pyparsing.infixNotation."""
        precedence_list = []
        for symbol_list in self.__ordered_operators:
            expression = Or(symbol_list)
            if type(self.__operators[symbol_list[0]]) is UnaryOperator:
                arity = 1
                parsing_function = parse_unary
                associativity = opAssoc.RIGHT
            else:
                arity = 2
                parsing_function = parse_binary
                associativity = opAssoc.LEFT

            entry = (expression, arity, associativity, parsing_function)
            precedence_list.append(entry)

        return precedence_list

    def __insert_operation(self, symbol: str, operation: Operation) -> None:
        """
        Insert an operation associated with a given symbol, updating
        __operators and __ordered_operators
        """
        if symbol in self.__operators:
            assert(isinstance(operation,
                              type(self.__operators[symbol].operations[-1])))
            self.__operators[symbol].operations.append(operation)
        else:
            if self.__growing_precedence:
                self.__ordered_operators.append([symbol])
            else:
                self.__ordered_operators[-1].append(symbol)

            operator = Operator.from_operation(symbol, operation)
            self.__operators[symbol] = operator

    def add_unary_operation(self, symbol: str, function: Callable[[Any], Any],
                            arg_type: Type, rtype: Optional[Type] = None) \
            -> None:
        """
        Add a unary operation to the map. Be carefull of the call order of this
        method: operators are added with decreasing precedence.
        """
        assert(len(signature(function).parameters) == 1)

        if rtype is None:
            rtype = arg_type

        operation = UnaryOperation(function, arg_type, rtype)
        self.__insert_operation(symbol, operation)

    def add_binary_operation(self, symbol: str,
                             function: Callable[[Any, Any], Any],
                             ltype: Type, rtype: Optional[Type] = None,
                             return_type: Optional[Type] = None) -> None:
        """
        Add a binary operation to the map. Be carefull of the call order of
        this method: operators are added with decreasing precedence.
        """
        assert(len(signature(function).parameters) == 2)

        if return_type is None:
            return_type = ltype

        if rtype is None:
            rtype = ltype

        operation = BinaryOperation(function, ltype, rtype, return_type)
        self.__insert_operation(symbol, operation)

    def __getitem__(self, symbol: str) -> Operator:
        """Return the BinaryOperator corresponding to the given symbol."""
        return self.__operators[symbol]

    def __lshift__(self, operation_tuple: Tuple[Any, ...]) -> None:
        """
        Helper function to call add_operation. Pass it a tuple with the same
        arguments you would pass to add_operation, in the same order. There
        is one exception : you can specify only the return type and left
        argument type of a binary operator, with (symbol, function, left,
        return).
        """
        assert(len(operation_tuple) >= 3)
        symbol = operation_tuple[0]
        function = operation_tuple[1]
        ltype = operation_tuple[2]

        arity = len(signature(function).parameters)
        assert(arity in (1, 2))

        if arity == 1:
            assert(len(operation_tuple) in (3, 4))
            self.add_unary_operation(*operation_tuple)
        elif arity == 2:
            assert(len(operation_tuple) in (3, 4, 5))

            if len(operation_tuple) == 4:
                return_type = operation_tuple[3]
                self.add_binary_operation(symbol, function, ltype,
                                          return_type=return_type)
            else:
                self.add_binary_operation(*operation_tuple)


OPERATORS_MAP = OperatorsMap()

# Arithmetic operators
OPERATORS_MAP << ('^', lambda l, r: l.value ** r.value, Type.Integer)

OPERATORS_MAP.start_precedence_group()
OPERATORS_MAP << ('*', lambda l, r: l.value * r.value, Type.Integer)
OPERATORS_MAP << ('/', lambda l, r: l.value / r.value, Type.Integer)
OPERATORS_MAP.end_precedence_group()

OPERATORS_MAP << ('\\', lambda l, r: l.value // r.value, Type.Integer)
OPERATORS_MAP << ('Mod', lambda l, r: l.value % r.value, Type.Integer)

OPERATORS_MAP.start_precedence_group()
OPERATORS_MAP << ('+', lambda l, r: l.value + r.value, Type.Integer)
OPERATORS_MAP << ('+', lambda l, r: l.value + r.value, Type.String)
OPERATORS_MAP << ('-', lambda l, r: l.value - r.value, Type.Integer)
OPERATORS_MAP.end_precedence_group()

# Concatenation operator
OPERATORS_MAP << ('&', lambda l, r: l.value + r.value, Type.String)

# Relational operators
OPERATORS_MAP.start_precedence_group()
OPERATORS_MAP << ('=', lambda l, r: l.value == r.value,
                  Type.Integer, Type.Boolean)
OPERATORS_MAP << ('=', lambda l, r: l.value == r.value,
                  Type.String, Type.Boolean)
OPERATORS_MAP << ('<>', lambda l, r: l.value != r.value,
                  Type.Integer, Type.Boolean)
OPERATORS_MAP << ('<>', lambda l, r: l.value != r.value,
                  Type.String, Type.Boolean)
OPERATORS_MAP << ('><', lambda l, r: l.value != r.value,
                  Type.Integer, Type.Boolean)
OPERATORS_MAP << ('><', lambda l, r: l.value != r.value,
                  Type.String, Type.Boolean)
OPERATORS_MAP << ('<', lambda l, r: l.value < r.value,
                  Type.Integer, Type.Boolean)
OPERATORS_MAP << ('>', lambda l, r: l.value > r.value,
                  Type.Integer, Type.Boolean)
OPERATORS_MAP << ('<=', lambda l, r: l.value <= r.value,
                  Type.Integer, Type.Boolean)
OPERATORS_MAP << ('>=', lambda l, r: l.value >= r.value,
                  Type.Integer, Type.Boolean)
OPERATORS_MAP.end_precedence_group()


# Logical and bitwise operators
OPERATORS_MAP << ('Not', lambda a: not a.value, Type.Boolean)
OPERATORS_MAP << ('And', lambda l, r: l.value and r.value, Type.Boolean)
OPERATORS_MAP << ('Or', lambda l, r: l.value or r.value, Type.Boolean)
OPERATORS_MAP << ('Xor', lambda l, r: l.value != r.value, Type.Boolean)
OPERATORS_MAP << ('Eqv', lambda l, r: l.value == r.value, Type.Boolean)
OPERATORS_MAP << ('Imp', lambda l, r: (not l.value) or r.value, Type.Boolean)



from enum import Enum
from typing import Optional, List, Tuple

from .error import CompilationError, ResolutionError


class Symbol:
    """
    A symbol extracted during compilation, representing a named programming
    element, and its hierarchical children.
    """
    Type = Enum("Type", "Namespace Module Function Variable")

    _name: str
    _symbol_type: Type
    _parent: Optional["Symbol"]
    _children: List["Symbol"]
    _position: Tuple[int]
    __search_nodes: List["Symbol"]
    __full_name: str

    @staticmethod
    def build_root() -> "Symbol":
        """Build a root namespace symbol, with no parent."""
        name = "<Root>"
        symbol_type = Symbol.Type.Namespace
        parent = None
        children: List['Symbol'] = []
        position: Tuple[int, ...] = tuple()
        root = Symbol(name, symbol_type, parent, children, position)
        root.__full_name = "<Root>"
        return root

    def __init__(self, name, symbol_type, parent=None, children=None,
                 position=None) -> None:
        """
        Shall not be called outside of Symbol or a child class. Use add_child
        instead.
        """
        self._name = name
        self._symbol_type = symbol_type
        self._parent = parent
        self._children = children
        self._position = position
        self.__full_name = ""

    @property
    def name(self):
        return self._name

    @property
    def symbol_type(self):
        return self._symbol_type

    @property
    def parent(self):
        return self._parent

    def __add_child(self, child) -> "Symbol":
        """Add an already initialized child."""
        if self.find_child(child._name) is None:
            child._parent = self
            child._position = self._position + (len(self._children), )
            child.__full_name = f"{self.__full_name}.{child._name}"
            self._children.append(child)
            return child
        else:
            msg = f"Trying to add the symbol {child._name} in {self}, which " \
                "already contains a symbol with the same name"
            raise CompilationError(msg)

    def add_child(self, name, child_type) -> "Symbol":
        """Initialize a child, add it and return it."""
        parent = None
        children: List['Symbol'] = []
        position = None
        child = Symbol(
            name, child_type, parent, children, position)
        return self.__add_child(child)

    def find_child(self, name: str) -> Optional["Symbol"]:
        """
        Search for a child with a given name. Return it or None if it is not
        found, and raise a CompilationError if there are two children with the
        same name.
        """

        results = list(
            filter(lambda child: child._name == name, self._children))
        if len(results) == 0:
            return None
        elif len(results) == 1:
            return results[0]
        else:
            msg = f"{self} has multiple children name {name}"
            raise CompilationError(msg)

    def find(self, name: str) -> List["Symbol"]:
        """
        Search for a direct or indirect child with the given name, using
        find_child recursively.
        """
        own_child = self.find_child(name)
        if own_child is None:
            ret = []
        else:
            ret = [own_child]

        return sum((child.find(name) for child in self._children), ret)

    def resolve(self, name: str) -> "Symbol":
        """Resolve a name in the recursive children of the symbol."""
        resolution = self.__resolve(name.split('.'))
        if resolution is None:
            msg = f"Can't resolve {name} from {self.full_name()}"
            raise ResolutionError(msg)
        else:
            return resolution

    def __resolve(self, name: List[str]) -> Optional["Symbol"]:
        if self._symbol_type == Symbol.Type.Function:
            if len(name) == 0:
                return self
            else:
                if self._parent is None:
                    return None
                else:
                    return self._parent.__resolve(name)
        else:
            child = self.find_child(name[0])
            if child is None:
                if self._parent is None:
                    return None
                else:
                    return self._parent.__resolve(name)
            else:
                return child.__resolve(name[1:])

    def is_root(self) -> bool:
        return self._parent is None

    def full_name(self) -> str:
        """Return the absolute name of the symbol, starting from the root"""
        return self.__full_name

    def __str__(self) -> str:
        return f"{self.full_name()}({self._symbol_type.name})"

    def __iter__(self):
        """
        Return an iterator to all the direct and indirect children of the
        symbol.
        """
        self.__search_nodes = [self]
        return self

    def __next__(self):
        try:
            last_node = self.__search_nodes.pop()
            self.__search_nodes.extend(last_node._children)
            return last_node
        except IndexError:
            raise StopIteration

    def __contains__(self, name):
        """Check if name is a direct child of the symbol."""
        return name in map(lambda s: s._name, self._children)



class Filesystem:
    variables = []

    @staticmethod
    def FolderExists(interpreter, arguments):
        interpreter.add_file_event('FolderExists', arguments[1].value)
        return False

    @staticmethod
    def CreateTextFile(interpreter, arguments):
        path = arguments[1].value
        interpreter.add_file_event('CreateTextFile', path)
        file_handler = interpreter.create_object('VBAEnv.WSH.FileHandler')
        file_handler.value['path'] = path
        return file_handler

    @staticmethod
    def CreateFolder(interpreter, arguments):
        path = arguments[1].value
        interpreter.add_file_event('CreateFolder', path)



"""Definition of the nodes of an abstract syntax tree."""

from abc import ABC, abstractmethod
from typing import List, Optional, Dict, Any, Union

from .operator import BinaryOperator
from .type import Type
from .visitor import Visitable


##################
#  Base classes  #
##################

class AST(Visitable, ABC):
    """Base class of all the nodes of the tree."""
    __ast_nodes_number: int = 0
    __hash: int

    @abstractmethod
    def __init__(self) -> None:
        self.__hash = AST.__ast_nodes_number
        AST.__ast_nodes_number += 1

    def to_dict(self) -> Dict[str, Any]:
        """
        Convert a tree to a recursive dict, using all the public members of
        the nodes as keys. A node has a '_type' key containing its type.
        """
        d: Dict[str, Any]

        def attribute_filter(attribute: str) -> bool:
            return not attribute.startswith('_') \
                and not callable(getattr(self, attribute))

        d = {'_type': type(self).__name__}
        for attr_name in filter(attribute_filter, dir(self)):
            attr = getattr(self, attr_name)

            if isinstance(attr, AST):
                d[attr_name] = attr.to_dict()
            elif isinstance(attr, (list, tuple)):
                d[attr_name] = []
                for elt in attr:
                    try:
                        d[attr_name].append(elt.to_dict())
                    except AttributeError:
                        d[attr_name].append(elt)
            else:
                d[attr_name] = str(attr)

        return d

    def __hash__(self) -> int:
        return self.__hash


class Statement(AST):
    """
    Generic statement, usually corresponding to a simple instruction, or the
    declaration of e.g. a function.
    """
    file: str
    line_number: int

    @abstractmethod
    def __init__(self, file: str = "", line_number: int = 0) -> None:
        """ Initialize a statement, saving its context. Must be called by
        child classes.

        :arg file: Name of the source file
        :arg line_number: Line number of the statement
        """
        super().__init__()
        self.file = file
        self.line_number = line_number


class Block(AST):
    """
    Sequence of statements, used as is to describe the main scope of a file,
    or inherited to implement block statements (loops, conditionals,
    functions definition, ...).
    """
    # TODO add support for file and line_number
    body: Optional[List[Union[Statement, "Block"]]]

    def __init__(self, body: Optional[List[Union[Statement, "Block"]]] = None)\
            -> None:
        super().__init__()
        self.body = body if body is not None else []


############################
#  Declarative statements  #
############################


class VarDec(Statement):
    """
    Single variable declaration, corresponding to a Dim or Const statement
    with a one-member variable list.
    """
    identifier: "Identifier"
    type: Optional[Union[Type, "Identifier"]]
    value: Optional["Expression"]
    new: bool

    def __init__(self, identifier: "Identifier",
                 type: Optional[Union[Type, "Identifier"]] = None,
                 value: Optional["Expression"] = None,
                 new: bool = False,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        self.identifier = identifier
        self.type = type
        self.value = value
        self.new = new


# TODO implement
class MultipleVarDec(Statement):
    """
    Variable declaration corresponding to a generic Dim or Const statement
    with potentially multiple variables.
    """
    declarations: List[VarDec]

    def __init__(self, declarations: List[VarDec], **kwargs) -> None:
        super().__init__(**kwargs)
        self.declarations = declarations


class VarAssign(Statement):
    """Variable assignment."""
    variable: Union["Get", "Identifier"]
    value: "Expression"

    def __init__(self, variable: "Get", value: "Expression",
                 **kwargs) -> None:
        super().__init__(**kwargs)
        self.variable = variable
        self.value = value


class FunDef(Block):
    """Function definition, corresponding to the Function keyword."""
    name: "Identifier"
    arguments: "ArgListDef"

    def __init__(self, name: "Identifier", arguments: "ArgListDef", **kwargs) \
            -> None:
        super().__init__(**kwargs)
        self.name = name
        self.arguments = arguments


class ProcDef(Block):
    """Procedure definition, corresponding to the Sub keyword."""
    name: "Identifier"
    arguments: "ArgListDef"

    def __init__(self, name: "Identifier", arguments: "ArgListDef", **kwargs) \
            -> None:
        super().__init__(**kwargs)
        self.name = name
        self.arguments = arguments


###########################
#  Executable statements  #
###########################


class Expression(Statement):
    """
    Root of an expression tree, made of binary and unary operators, with
    literals, identifiers and function calls as leafs.
    """

    @abstractmethod
    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)


class Identifier(Expression):
    """Identifier of a variable or function."""
    name: str

    def __init__(self, name: str, **kwargs) -> None:
        super().__init__(**kwargs)
        self.name = name


class Get(Expression):
    """Recursive node corresponding to the . operator."""
    parent: Union["Get", Identifier, "FunCall"]
    child: Identifier

    def __init__(self, parent: Union["Get", Identifier, "FunCall"],
                 child: Identifier, **kwargs) -> None:
        super().__init__(**kwargs)
        self.parent = parent
        self.child = child


class Literal(Expression):
    """Literal value : integer, double, boolean, string, ..."""
    type: Type
    value: Union[int, float, bool, str]

    def __init__(self, type: Type, value: Union[int, float, bool, str],
                 **kwargs) -> None:
        super().__init__(**kwargs)
        self.value = value
        self.type = type

    @staticmethod
    def from_value(value) -> "Literal":
        return Literal(value.base_type, value.value)


class ArgListCall(Statement):
    """List of arguments, used by function calls"""
    args: List["Expression"]

    def __init__(self, args: List["Expression"], **kwargs) -> None:
        super().__init__(**kwargs)
        self.args = args


class ArgListDef(Statement):
    """List of arguments, used by function declarations"""
    args: List[Identifier]

    def __init__(self, args: List["Identifier"], **kwargs) -> None:
        super().__init__(**kwargs)
        self.args = args


class FunCall(Expression):
    """Function call"""
    function: Union[Get, Identifier]
    arguments: ArgListCall

    def __init__(self, function: Union[Get, Identifier],
                 arguments: ArgListCall, **kwargs) -> None:
        super().__init__(**kwargs)
        self.function = function
        self.arguments = arguments


class UnOp(Expression):
    """Unary operator"""
    operator: str
    argument: Expression

    def __init__(self, operator: str, argument: Expression, **kwargs) -> None:
        super().__init__(**kwargs)
        self.operator = operator
        self.argument = argument


class BinOp(Expression):
    """Binary operator"""
    operator: str
    left: Expression
    right: Expression

    def __init__(self, operator: str, left: Expression,
                 right: Expression, **kwargs) -> None:
        super().__init__(**kwargs)
        self.operator = operator
        self.left = left
        self.right = right


##########################
#  Loop and conditional  #
##########################

class ElseIf(Block):
    """Single condition/action block, used internally by If"""
    condition: Expression

    def __init__(self, condition: Expression, **kwargs) -> None:
        super().__init__(**kwargs)
        self.condition = condition


class If(Block):
    """If statement"""
    condition: Expression
    elsifs: List[ElseIf]
    else_block: Optional[Block]

    def __init__(self, condition: Expression,
                 elsifs: Optional[List[ElseIf]] = None,
                 else_block: Optional[Block] = None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.condition = condition
        self.elsifs = elsifs if elsifs is not None else []
        self.else_block = else_block


class For(Block):
    """For statement with a counter"""
    counter: Identifier
    start: Expression
    end: Expression
    step: Optional[Expression]

    def __init__(self, counter: Identifier, start: Expression, end: Expression,
                 step: Optional[Expression] = None, **kwargs) -> None:
        super().__init__(**kwargs)
        self.counter = counter
        self.start = start
        self.end = end
        self.step = step

####################
#  Error handling  #
####################


class OnError(Statement):
    """
    On Error statement. If goto is None, corresponds to a Resume Next policy,
    else to a GoTo policy.
    """
    goto: Optional[Union[Literal, Identifier]]

    def __init__(self, goto: Optional[Union[int, Identifier]] = None,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        self.goto = goto


class Resume(Statement):
    """
    Resume statement. If goto is None, corresponds to Resume or Resume Next,
    else to a GoTo form.
    """
    goto: Optional[Union[Literal, Identifier]]

    def __init__(self, goto: Optional[Union[int, Identifier]] = None,
                 **kwargs) -> None:
        super().__init__(**kwargs)
        self.goto = goto


class ErrorStatement(Statement):
    """Error statement."""
    number: Literal

    def __init__(self, number: Literal, **kwargs) -> None:
        super().__init__(**kwargs)
        self.number = number




class PreprocessorError(Exception):
    """
    Error raised during preprocessing.
    """
    file_name: str
    line_number: int
    message: str

    def __init__(self, file_name: str, line_number: int, message: str) -> None:
        self.file_name = file_name
        self.line_number = line_number
        self.message = message

    def __str__(self) -> str:
        return f"{self.file_name}:{self.line_number}: {self.message}"


class ParsingError(Exception):
    """
    Error raised during parsing.
    """

    def __init__(self, file_name: str, line_number: int, message: str):
        self.file_name = file_name
        self.line_number = line_number
        self.message = message

    def __str__(self) -> str:
        return f"{self.file_name}:{self.line_number}: {self.message}"


class InterpretationError(Exception):
    pass


class CompilationError(Exception):
    pass


class OperatorError(Exception):
    pass


class ConversionError(Exception):
    pass


class ResolutionError(Exception):
    pass


class SerializationError(Exception):
    pass



#! /usr/bin/env python

from argparse import ArgumentParser
from pathlib import Path
from sys import exit

from magic import from_buffer as magic_from_buffer

from oletools.olevba import VBA_Parser

from emu import (Program, Compiler, Interpreter, Unit, __version__,
                 OutsideWorld, ReportGenerator, Serializer, SerializationError)


def build_argparser():
    parser = ArgumentParser(
        description="""VBA static and dynamic analysis tool for malware
                       analysts."""
    )
    parser.add_argument(
        "-v", "--version",
        action='version',
        version=f"SpuriousEmu v{__version__}")
    subparsers = parser.add_subparsers(
        dest="mode",
        title="Operating mode",
        required=True)

    # Static analysis
    static_parser = subparsers.add_parser(
        'static',
        help="Static analysis, allows symbols preview, compilation and "
             "deobfuscation.")
    static_parser.add_argument(
        '-o', '--output',
        help="Output file, used to save the compilation result or the "
             "deobfuscated macros")
    static_parser.add_argument(
        "-d", "--deobfuscate",
        metavar="LEVEL",
        type=int,
        help="Enable deobfuscation and specify the deobfuscation level. See "
             "-e to only process a specific symbol")
    static_parser.add_argument(
        "-e", "--entry",
        help="Absolute path of the class or function to deobfuscate")
    static_parser.add_argument(
        "input",
        help="""Input file, can be an Office document, some VBA source code, or
                a SpuriousEmu compiled program""")
    static_parser.set_defaults(func=static_analysis)

    # Dynamic analysis
    dynamic_parser = subparsers.add_parser(
        'dynamic',
        help="Dynamic analysis used to perform different kind of emulation.")
    dynamic_parser.add_argument(
        "-e", "--entry",
        default='Main',
        help="""Entry point of the dynamic analysis. Must be specified if no
                Main function is defined. Must use the absolute path of the
                symbol""")
    dynamic_parser.add_argument(
        "-r", "--results",
        help="Save the analysis results to REPORT")
    dynamic_parser.add_argument(
        "-o", "--output",
        help="Directory to save the created files to")
    dynamic_parser.add_argument(
        "input",
        help="""Input file, can be an Office document, some VBA source code, or
                a SpuriousEmu compiled program""")
    dynamic_parser.set_defaults(func=dynamic_analysis)

    # Report generation
    report_parser = subparsers.add_parser(
        'report',
        help="""Extract information from analysis results generated by other
                commands."""
    )
    report_action = report_parser.add_mutually_exclusive_group(required=True)
    report_action.add_argument(
        '-y', '--symbols',
        action="store_true",
        help="Display the list of functions and classes of a .spemu-com file")
    report_action.add_argument(
        "-x", "--extract-files",
        metavar="DIR",
        help="""Extract files from a dynamic analysis report to DIR. Defaults
                to --table""")
    report_action.add_argument(
        "-t", "--timeline",
        action="store_true",
        help="Display the event timeline. Defaults to --table")
    report_action.add_argument(
        "-c", "--categories",
        action="store_true",
        help="Display all the events, grouped by category. Defaults to --json")
    report_format = report_parser.add_mutually_exclusive_group()
    report_format.add_argument(
        "--json",
        action="store_true",
        help="Use JSON output")
    report_format.add_argument(
        "--csv",
        action="store_true",
        help="Use CSV output")
    report_format.add_argument(
        "--table",
        action="store_true",
        help="Use human-friendly output")
    report_parser.add_argument(
        '-s', '--shorten',
        action="store_true",
        help="Used with --table, produces shorter tables")
    report_parser.add_argument(
        '-k', '--skip-streaks',
        type=int,
        metavar="LENGTH",
        help="""Used with --shorten and --table, only display the first
                elements of a series of similar events""")
    report_parser.add_argument(
        "input",
        help="Dynamic result file to use")
    report_parser.set_defaults(func=generate_report)

    return parser


def compile_input_file(input_file: str) -> Program:
    path = Path(input_file)

    if path.is_dir():
        # A directory must be a project containing VBA source files
        return Compiler.compile_project(input_file)

    with open(input_file, 'rb') as f:
        content = f.read()

    # Try to deserialize an already compiled program
    try:
        serializer = Serializer()
        obj = serializer.deserialize(content)

        if type(obj) == Program:
            return obj
    except SerializationError:
        pass

    # Compile an Office document or VBA source files
    units = []
    if magic_from_buffer(content, mime=True) == "text/plain":
        if not input_file.endswith('.vbs'):
            input_file += '.vbs'
        units.append(Unit.from_content(content.decode('utf-8'), input_file))
    else:
        vba_parser = VBA_Parser(input_file, data=content)

        for _, _, vba_filename, vba_code in vba_parser.extract_all_macros():
            units.append(Unit.from_content(vba_code, vba_filename))

    return Compiler.compile_units(units)


def static_analysis(arguments):
    # Check arguments validity
    if not Path(arguments.input).exists():
        print(f"Error: input file {arguments.input} does not exist.")
        return 1

    if arguments.deobfuscate is not None:
        print(f"Error: de-obfuscation is not handled yet.")
        return 1

    if (arguments.entry is not None) and arguments.deobfuscate is None:
        print(f"Error: the -e flag requires the -d flag.")
        return 1

    # Compile program
    program = compile_input_file(arguments.input)

    # Display symbols
    report_generator = ReportGenerator(program=program)
    report_generator.output_format = ReportGenerator.Format.TABLE
    print(report_generator.produce_symbols())

    if arguments.output is not None:
        report_generator.save_program(arguments.output)

    return 0


def execute_program(program: Program, entry_point: str) -> None:
    libraries = list(map(str, Path('./lib/').glob('*')))
    linked_program = Compiler.link_libraries(program, libraries)
    return Interpreter.run_program(linked_program, entry_point)


def dynamic_analysis(arguments):
    # Check arguments validity
    if not Path(arguments.input).exists():
        print(f"Error: input file {arguments.input} does not exist.")
        return 1

    # Load and execute
    program = compile_input_file(arguments.input)
    outside_world = execute_program(program, arguments.entry)

    # Produce timeline table report
    report_generator = ReportGenerator(outside_world=outside_world)
    report_generator.output_format = report_generator.Format.TABLE
    report_generator.shorten = True
    report_generator.skip_similar = 3
    print(report_generator.produce_timeline())

    # Export whole analysis
    if arguments.results is not None:
        report_generator.save_outside_world(arguments.results)

    # Save created files
    if arguments.output is not None:
        report_generator.extract_files(arguments.output)

    return 0


def generate_report(arguments):
    # Check input
    if not Path(arguments.input).exists():
        print(f"Error: input file {arguments.input} does not exist.")
        return 1

    # Prepare report generator
    result = Serializer.load(arguments.input)

    if isinstance(result, Program):
        report_generator = ReportGenerator(program=result)
    elif isinstance(result, OutsideWorld):
        report_generator = ReportGenerator(outside_world=result)
    else:
        print("Report generation is not yet supported for type "
              f"{type(result)}")

    format_specified = True
    if arguments.json:
        report_generator.output_format = report_generator.Format.JSON
    elif arguments.csv:
        report_generator.output_format = report_generator.Format.CSV
    elif arguments.table:
        report_generator.output_format = report_generator.Format.TABLE
    else:
        format_specified = False

    report_generator.shorten = arguments.shorten
    report_generator.skip_similar = arguments.skip_streaks

    # Produce report
    if arguments.symbols:
        if not format_specified:
            report_generator.output_format = report_generator.Format.TABLE

        print(report_generator.produce_symbols())

    if arguments.extract_files is not None:
        report_generator.extract_files(arguments.extract_files)

    if arguments.timeline:
        if not format_specified:
            report_generator.output_format = report_generator.Format.TABLE

        print(report_generator.produce_timeline())

    if arguments.categories:
        if not format_specified:
            report_generator.output_format = report_generator.Format.JSON

        print(report_generator.produce_organized_events())

    return 1


def main():
    parser = build_argparser()
    args = parser.parse_args()

    return args.func(args)


if __name__ == "__main__":
    exit(main())



"""
Implement the different kind of Reference classes produced by the compilation
"""

from abc import ABC, abstractmethod
from enum import Enum
from typing import Any, Dict, List, Optional

from .error import ResolutionError, CompilationError


class Reference(ABC):
    """
    Semantic representation of an identifier, either structural (for
    namespaces) or computational (for variables and functions). The references
    of a program are seen as a tree denoting its organization, with variables
    as the only kind of leaf.
    """
    Category = Enum("Category", "Structural Computational")

    category: "Reference.Category"

    name: str
    parent: Optional["Reference"]
    children: List["Reference"]
    __full_name: str

    @abstractmethod
    def __init__(self, name: str, parent: "Reference" = None,
                 children: List["Reference"] = None) -> None:
        self.name = name
        self.parent = parent
        self.children = children if children is not None else []
        self.__full_name = self.name

    def get_child(self, name: str) -> "Reference":
        """
        Get a child from the current reference, by its name.

        :throws ResolutionError: If there is no child with the given name
        """
        ret = list(filter(lambda child: child.name == name, self.children))
        assert(len(ret) <= 1)

        if len(ret) == 1:
            return ret[0]
        else:
            msg = f"Reference {self} doesn't have child called {name}"
            raise ResolutionError(msg)

    def add_child(self, child: "Reference") -> None:
        """
        Add a child to the current reference, updating its parent field.

        :throws CompilationError: If there is already a child with the same
        name
        """
        try:
            self.get_child(child.name)
            msg = f"Reference {self} already has a child called {child.name}"
            raise CompilationError(msg)
        except ResolutionError:
            self.children.append(child)
            child.parent = self
            child.__full_name = f"{self.__full_name}.{child.name}"

    def build_child(self, child_type, *args, **kwargs) -> "Reference":
        """
        Build a reference, add it as a child and return it.

        :param child_type: Type of the child to create
        :param args: Arguments to pass to the child constructor
        :param kwargs: Keyword arguments to pass to the child constructor
        :returns: The built child
        """
        child = child_type(*args, **kwargs)
        self.add_child(child)
        return child

    def search_child(self, name: str,
                     exclude_child: Optional["Reference"] = None) \
            -> "Reference":
        try:
            return self.get_child(name)
        except ResolutionError:
            for child in self.children:
                if child is exclude_child:
                    pass

                try:
                    return child.search_child(name)
                except ResolutionError:
                    pass

        msg = f"Can't find a reference named {name} in {self} children"
        raise ResolutionError(msg)

    def to_dict(self) -> Dict[str, Any]:
        d = {'name': self.name}
        for child in self.children:
            child_type = type(child).__name__
            similar_childs = d.get(child_type, [])
            d[child_type] = similar_childs + [child.to_dict()]

        return d

    def __str__(self) -> str:
        """Return the absolute name of the reference"""
        return self.__full_name


class Environment(Reference):
    category = Reference.Category.Structural

    DEFAULT_NAME = "VBAEnv"

    def __init__(self, name: Optional[str] = None) -> None:
        if name is None:
            name = Environment.DEFAULT_NAME

        super().__init__(name=name)


class Project(Reference):
    category = Reference.Category.Structural

    def __init__(self, name: str) -> None:
        super().__init__(name=name)


class Module(Reference):
    category = Reference.Category.Structural

    @abstractmethod
    def __init__(self, name: str) -> None:
        super().__init__(name=name)

    @property
    def functions(self):
        return list(filter(lambda child: type(child) is FunctionReference,
                           self.children))

    @property
    def variables(self):
        return list(filter(lambda child: type(child) is Variable,
                           self.children))

    def get_function(self, name: str) -> "FunctionReference":
        function = self.get_child(name)
        assert(type(function) is FunctionReference)
        return function

    def get_variable(self, name: str) -> "Variable":
        variable = self.get_child(name)
        assert(type(variable) is Variable)
        return variable


class ClassModule(Module):
    category = Reference.Category.Structural

    def __init__(self, name: str) -> None:
        super().__init__(name=name)


class ProceduralModule(Module):
    category = Reference.Category.Structural

    def __init__(self, name: str) -> None:
        super().__init__(name=name)


class Variable(Reference):
    class Extent(Enum):
        Program = 1
        Module = 2
        Procedure = 3
        Object = 4
        Aggregate = 5

    category = Reference.Category.Computational

    extent: "Variable.Extent"

    def __init__(self, name: str, extent: "Variable.Extent") -> None:
        super().__init__(name=name)
        self.extent = extent

    def add_child(self, child: Reference) -> None:
        msg = f"Can't add a child to Variable {self}"
        raise CompilationError(msg)

    def get_child(self, name: str) -> Reference:
        msg = f"Variable reference {self} can't have a child"
        raise ResolutionError(msg)


class FunctionReference(Reference):
    category = Reference.Category.Computational

    def __init__(self, name: str) -> None:
        super().__init__(name=name)



from dataclasses import dataclass
from inspect import getfullargspec
from typing import Callable, List, Optional, Union

from .abstract_syntax_tree import Block
from .reference import FunctionReference
from .value import Value, Object
from .type import Type


@dataclass
class Function(Value):
    """An abstract function, be it implemented in VBA or Python."""
    base_type = Type.Function

    name: str
    arguments_names: List[str]
    reference: Optional[FunctionReference]
    parent_object: Optional[Object]

    def __init__(self, name: str, arguments_names: List[str],
                 function: Union["ExternalFunction.Signature", Block],
                 reference: Optional[FunctionReference] = None) -> None:
        self.name = name
        self.arguments_names = arguments_names
        self.value = function
        self.reference = reference
        self.parent_object = None

    def convert_to_different_type(self, to_type: Type) -> Optional[Value]:
        return None

    def create_bound_method(self, parent_object: Object) -> "Function":
        method = type(self)(self.name, self.arguments_names, self.value,
                            self.reference)
        method.parent_object = parent_object
        return method


class ExternalFunction(Function):
    """
    Python function wrapper, the function will be called with two arguments:
      - a reference to the interpreter, allowing to have side effects
      - a list of Value objects, the actual arguments of the function
    It has to return a Value object.
    """
    Signature = Callable[["Interpreter", List[Value]], Value]

    def __init__(self, name: str, arguments_names: List[str],
                 external_function: "ExternalFunction.Signature",
                 reference: Optional[FunctionReference] = None) -> None:
        super().__init__(name, arguments_names, external_function, reference)

    @staticmethod
    def from_function(
            python_function: Optional["ExternalFunction.Signature"] = None,
            name: Optional[str] = None) -> \
        Union["ExternalFunction.Signature",
              Callable[["ExternalFunction.Signature"], "ExternalFunction"]]:
        """
        Create an ExternalFunction from a Python function, optionally changing
        its name.
        """
        def decorator(python_function: "ExternalFunction.Signature") \
                -> "ExternalFunction":
            new_name = python_function.__name__ if name is None else name
            arguments_names = list(getfullargspec(python_function)[0])
            return ExternalFunction(new_name, arguments_names, python_function)

        if python_function is None:
            return decorator
        else:
            return decorator(python_function)

    @property
    def external_function(self) -> "ExternalFunction.Signature":
        return self.value


class InternalFunction(Function):
    """VBA function."""

    def __init__(self, name: str, arguments_names: List[str],
                 body: Block, reference: FunctionReference) -> None:
        super().__init__(name, arguments_names, body, reference)

    @property
    def body(self) -> Block:
        return self.value



"""Define classes allowing to store object values"""

# See type.py for a list of type still to be implemented
#
# To add a new supported value, you must:
#  - subclass Value
#  - add the Type/Value correspondance in TYPES_MAP

from abc import abstractmethod, ABC
from typing import Any, Dict, Optional, Union

from .error import ConversionError
from .reference import ClassModule
from .type import Type


class Value(ABC):
    """Represents the interpretation of an expression."""
    base_type: Type
    value: Any

    @abstractmethod
    def __init__(self):
        pass

    def convert_to(self, to_type: Type) -> "Value":
        """Used to convert values between types, for example to get the
        boolean interpretation of a non Boolean value."""
        if self.base_type == to_type:
            return self
        else:
            converted = self.convert_to_different_type(to_type)
            if converted is not None:
                return converted
            else:
                msg = f"Can't convert {self.base_type} to {to_type}"
                raise ConversionError(msg)

    @abstractmethod
    def convert_to_different_type(self, to_type: Type) -> Optional["Value"]:
        """
        Function overridden by child classes, to define how to convert the
        value to different types. It should return None for non-supported
        types. Don't override convert_to !
        """
        pass

    @staticmethod
    def from_value(value, to_type: Type) -> "Value":
        """
        Use a couple value/type to build a Value subclass object.

        >>> str(Value.from_value(12, Type.Integer))
        'Integer(12)'

        It relies on the TYPES_MAP dictionnary.
        """
        try:
            return TYPES_MAP[to_type](value)
        except KeyError:
            raise NotImplementedError(f"Type {to_type} is not already "
                                      "implemented")

    @staticmethod
    def from_literal(literal) -> "Value":
        """
        Build the value corresponding to a literal, using Value.from_value.
        """
        return Value.from_value(literal.value, literal.type)

    def __str__(self) -> str:
        return f"{self.base_type.name}({self.value})"

    @staticmethod
    def from_python_base_type(value: Any) -> "Value":
        value_type = type(value)
        if isinstance(value, Value) or value is None:
            return value
        elif value_type is int:
            return Integer(value)
        elif value_type is bool:
            return Boolean(value)
        elif value_type is str:
            return String(value)
        else:
            msg = f"Can't create VBA value for Python type {value_type}"
            raise ConversionError(msg)


class Integer(Value):
    base_type = Type.Integer

    def __init__(self, integer: Union[str, int]) -> None:
        if isinstance(integer, int):
            self.value = integer
        else:
            self.value = int(integer, 10)

    def convert_to_different_type(self, to_type: Type) -> Optional[Value]:
        if to_type == Type.Boolean:
            return Boolean(self.value != 0)
        return None


class Boolean(Value):
    base_type = Type.Boolean

    def __init__(self, boolean: Union[str, bool]) -> None:
        if isinstance(boolean, bool):
            self.value = boolean
        else:
            self.value = boolean == "True"

    def convert_to_different_type(self, to_type: Type) -> Optional[Value]:
        if to_type == Type.Integer:
            return Integer(1) if self.value else Integer(0)
        return None


class String(Value):
    base_type = Type.String

    def __init__(self, string: str) -> None:
        self.value = string

    def convert_to_different_type(self, to_type: Type) -> Optional[Value]:
        # TODO
        return None


class Object(Value):
    base_type = Type.Object

    class_reference: ClassModule
    value: Dict[str, Value]  # Holds the state of the object

    def __init__(self, class_reference=ClassModule,
                 variables: Dict[str, Value] = None) -> None:
        self.class_reference = class_reference
        self.value = variables if variables is not None else dict()

    @property
    def variables(self) -> Dict[str, Value]:
        return self.value

    def convert_to_different_type(self, to_type: Type) -> Optional[Value]:
        # TODO
        return None


TYPES_MAP = {
    Type.Integer: Integer,
    Type.Boolean: Boolean,
    Type.String: String
}



class WshShell:
    variables = []

    @staticmethod
    def Run(interpreter, arguments):
        # TODO handle other arguments
        command = arguments[1].value
        interpreter.add_command_execution(command)



"""Tools to generate reports based on analysis results."""

import csv
import json

from dataclasses import dataclass, field
from enum import Enum
from io import StringIO
from hashlib import md5
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple

from prettytable import PrettyTable

from .compiler import Program
from .side_effect import OutsideWorld
from .serialize import Serializer


def _needs_outside_world(method):
    error_msg = "ReportGenerator needs an OutsideWorld object to call " \
        f"{method.__name__}"

    def decorated(*args, **kwargs):
        if args[0].outside_world is None:
            raise RuntimeError(error_msg)

        return method(*args, **kwargs)

    return decorated


def _needs_program(method):
    error_msg = "ReportGenerator needs a Program object to call " \
        f"{method.__name__}"

    def decorated(*args, **kwargs):
        if args[0].program is None:
            raise RuntimeError(error_msg)

        return method(*args, **kwargs)

    return decorated


@dataclass
class ReportGenerator:
    """
    Class used to extract information from an OutsideWorld object. You can
    get different kind of reports (timeline of the events, hierarchical output
    of all the events, ...) using different formats, specified in the Format
    enumeration. You can also use it to extract generated files and save them
    to you file system.

    You can customize the generated reports using the object attributes:
        - output_format : the output format
        - indent : Number of spaces for JSON indentation
        - reproducible: If True, don't display event time
        - skip_similar: If > 0, skip series of similar events in TABLE output
        - hash_algorithm: hashing algorithm used to name files
        - shorten: If True, shorten the context and data fields in TABLE output
    """
    class Format(Enum):
        """Display formats supported by the ReportGenerator."""
        JSON = 'json'
        CSV = 'csv'
        TABLE = 'table'

    EventLine = Tuple[int, str, str, str, Any]

    program: Optional[Program] = None
    outside_world: Optional[OutsideWorld] = None
    output_format: "ReportGenerator.Format" = field(default=Format.JSON)
    indent: int = 4
    reproducible: bool = False
    skip_similar: int = 0
    hash_algorithm = md5
    shorten = False

    # Utility methods

    def hash_file(self, content: str) -> str:
        """
        Return the hex digest of the file content encoded in UTF8, using
        the self.hash algorithm.
        """
        hasher = self.hash_algorithm()
        hasher.update(content.encode('utf-8'))

        return hasher.hexdigest()

    def to_json(self, report: Any) -> str:
        """Return the JSON dump of the report."""
        return json.dumps(report, indent=self.indent, sort_keys=True)

    def event_to_tuple(self, event: OutsideWorld.Event) \
            -> "ReportGenerator.EventLine":
        """
        Convert an event to a tuple.
        """
        identifier = event.identifier
        time = f"{event.time:.3f}"
        category = event.category.value
        context = event.context
        data = event.data

        return (identifier, time, category, context, data)

    def shorten_tuple(self, event: "ReportGenerator.EventLine") \
            -> "ReportGenerator.EventLine":
        """
        Shorten a tuple if self.shorten is True, ie. only keep context end
        symbol and discard data['data'] if possible.
        """
        if not self.shorten:
            return event

        identifier = event[0]
        time = event[1]
        category = event[2]
        context = event[3].split('.')[-1]

        if isinstance(event[4], dict):
            data = {key: value for key, value in event[4].items()
                    if key != "data"}
        else:
            data = event[4]

        return (identifier, time, category, context, data)

    def similar_events(self, event1: "ReportGenerator.EventLine",
                       event2: "ReportGenerator.EventLine") -> bool:
        """
        Compare two events, looking for similarities. Two events are similar
        if:
            - they have the same category and
            - they have the same context
            - they have dict data field and
                - they have data['data'] and data['type'] fields
                - their data['type'] fields are the same
        """
        category1 = event1[2]
        category2 = event2[2]
        if category1 != category2:
            return False

        context1 = event1[3]
        context2 = event2[3]
        if context1 != context2:
            return False

        data1 = event1[4]
        data2 = event2[4]

        if not isinstance(data1, dict):
            return True

        if 'data' not in data1 or 'data' not in data2:
            return False

        if 'type' not in data1 or 'type' not in data2:
            return False

        return data1['type'] == data2['type']

    def events_to_table(self, events: List[OutsideWorld.Event]) -> PrettyTable:
        """
        Build a table containing a list of events. The table has the columns
        ('ID', 'Time (s)', 'Category', 'Context', 'Data') and an event by row.

        Depending on shorten and skip_similar, series of similar events can be
        skiped.
        """
        fields = ('ID', 'Time (s)', 'Category', 'Context', 'Data')
        table = PrettyTable(fields)
        table.align = 'l'

        similar_events_streak = 0
        previous_event_tuple = None

        for event_tuple in events:
            if previous_event_tuple is not None:
                if self.similar_events(previous_event_tuple, event_tuple):
                    similar_events_streak += 1
                else:
                    if similar_events_streak > self.skip_similar + 1 \
                       and self.shorten:
                        table.add_row(('...', ) * 5)

                    if similar_events_streak > 1:
                        table.add_row(
                            self.shorten_tuple(previous_event_tuple))
                    similar_events_streak = 0

            if similar_events_streak <= self.skip_similar \
               or not self.shorten:
                table.add_row(self.shorten_tuple(event_tuple))

            previous_event_tuple = event_tuple

        if similar_events_streak > self.skip_similar + 1 \
           and self.shorten:
            table.add_row(('...', ) * 5)

        if similar_events_streak > 1:
            table.add_row(
                self.shorten_tuple(previous_event_tuple))

        return table

    def events_to_csv(self, events: List["ReportGenerator.EventLine"]) -> str:
        """Return a CSV formatted representation of a list of events."""
        stream = StringIO()
        writer = csv.writer(stream,
                            delimiter=";",
                            quoting=csv.QUOTE_NONNUMERIC)

        writer.writerow(('ID', 'Time (s)', 'Category', 'Context', 'Data'))

        for event in events:
            writer.writerow(event)

        return stream.getvalue()

    # Program reports

    @_needs_program
    def extract_symbols(self) -> Dict[str, List[str]]:
        """Returns a dictionnary with 'functions' and 'classes' keys."""
        memory_dict = self.program.to_dict()['memory']
        functions = memory_dict['functions']
        classes = memory_dict['classes']

        return {'functions': functions, 'classes': classes}

    @_needs_program
    def produce_symbols(self) -> str:
        """Returns the formatted symbols : classes and functions."""
        symbols = self.extract_symbols()

        if self.output_format is ReportGenerator.Format.JSON:
            return self.to_json(symbols)
        if self.output_format is ReportGenerator.Format.CSV:
            return "CSV not supported"
        elif self.output_format is ReportGenerator.Format.TABLE:
            classes = PrettyTable()
            classes.add_column("Classes", symbols['classes'], align="l")

            functions = PrettyTable()
            functions.add_column("Functions", symbols['functions'], align='l')

            return f"{classes}\n\n{functions}"

    @_needs_program
    def save_program(self, save_path: str) -> None:
        """Save the program to save_path."""
        Serializer.save(self.program, save_path)

    # OutsideWorld reports

    # TODO return a list of Events, to ease JSON output
    @_needs_outside_world
    def extract_timeline(self) -> List["ReportGenerator.EventLine"]:
        """
        Return the chronological list of events, as tuples with elements
        identifier, time, category, context and data.
        """

        timeline = list(map(self.event_to_tuple, self.outside_world.events))
        timeline.sort()

        return timeline

    @_needs_outside_world
    def produce_timeline(self) -> str:
        """Return the formatted timeline."""
        timeline = self.extract_timeline()

        if self.output_format is ReportGenerator.Format.JSON:
            return json.dumps(timeline, indent=self.indent, sort_keys=True)
        elif self.output_format is ReportGenerator.Format.CSV:
            return self.events_to_csv(timeline)
        elif self.output_format is ReportGenerator.Format.TABLE:
            table = self.events_to_table(timeline)

            if self.shorten:
                return table.get_string(
                    fields=('ID', 'Category', 'Context', 'Data'))

            return table.get_string()

    @_needs_outside_world
    def produce_organized_events(self) -> str:
        """Return the events, organized by category."""
        report = dict()

        for event in self.outside_world.events:
            category = event.category.value
            event_dict = {
                'identifier': event.identifier,
                'time': event.time,
                'context': event.context,
                'data': event.data
            }

            category_list = report.get(event.category.value, [])
            report[category] = category_list + [event_dict]

        if self.output_format == ReportGenerator.Format.JSON:
            return self.to_json(report)
        elif self.output_format == ReportGenerator.Format.CSV:
            return "CSV not supported yet"
        elif self.output_format == ReportGenerator.Format.TABLE:
            output = ""

            for category in report:
                output += f"{category}:\n"
                events = [self.event_to_tuple(event)
                          for event in self.outside_world.events
                          if event.category.value == category]
                table = self.events_to_table(events)

                if self.shorten:
                    fields = ('ID', 'Context', 'Data')
                else:
                    fields = ('ID', 'Time (s)', 'Context', 'Data')

                output += table.get_string(fields=fields)
                output += "\n\n"

            return output

    @_needs_outside_world
    def extract_files(self, save_directory: str) -> None:
        """
        Extract the content added to the OutsideWorld files, and write it to
        the actual filesystem. It uses the following convention: for each
        file that exists in OutsideWorld, with md5 sum `hash`, write to
        `hash` the content of the file, and to `hash.filename.txt` its name. In
        case of filename conflict, overrides the file.

        :arg save_directory: Output directory, created if it does not exist.
        """
        path = Path(save_directory)
        path.mkdir(exist_ok=True)

        for name, content in self.outside_world.files.items():
            content_hash = self.hash_file(content)
            content_path = path.joinpath(content_hash)
            filename_path = path.joinpath(f'{content_hash}.filename.txt')

            with open(content_path.absolute(), 'w') as f:
                f.write(content)

            with open(filename_path.absolute(), 'w') as f:
                f.write(name)

    @_needs_outside_world
    def save_outside_world(self, save_path: str) -> None:
        """Save the outside world to save_path."""
        Serializer.save(self.outside_world, save_path)



class FileHandler:
    variables = ["path", "open"]

    @staticmethod
    def Write(interpreter, arguments):
        # TODO check open
        path = arguments[0].value['path']
        interpreter.add_file_event("Write", path, arguments[1].value)

    @staticmethod
    def Close(interpreter, arguments):
        # TODO set open to False
        pass



"""Allow serialization of SpuriousEmu objects"""

import pickle

from dataclasses import dataclass
from pathlib import Path
from typing import List, Type, Union

from .compiler import Program
from .error import SerializationError
from .side_effect import OutsideWorld


class Serializer:
    """
    Set of class methods allowing to serialize, deserialize, save and load
    SpuriousEmu objects.

    The specifications of the different supported formats, ie. their magic
    string, corresponding Python classes, and optional extensions, are stored
    in Serialize.FORMATS.
    """
    @dataclass
    class FileFormat:
        object_type: type
        magic_string: bytes
        extension: str

    FORMATS: List[FileFormat] = [
        FileFormat(Program, b'SpuriousEmuProgram', ".spemu-com"),
        FileFormat(OutsideWorld, b'SpuriousEmuOutsideWorld', ".spemu-out")
    ]

    SerializableType = Union[Program, OutsideWorld]

    @classmethod
    def serialize(cls, obj: SerializableType) -> bytes:
        if not isinstance(obj, Serializer.SerializableType.__args__):
            msg = f"Serialization of {type(obj)} is not supported"
            raise SerializationError(msg)

        body = pickle.dumps(obj)
        magic = cls.magic(type(obj))

        return magic + body

    @classmethod
    def save(cls, obj: SerializableType, file_path: str) -> None:
        """
        Serialize a Python object and save it to a file, potentially overriding
        it. If the save path has no extension, add the corresponding one.
        """
        path = Path(file_path)
        print(path.suffix)
        if path.suffix == "":
            save_path = path.parent / (path.name + cls.extension(type(obj)))
        else:
            save_path = path

        content = cls.serialize(obj)

        with open(save_path, 'wb') as f:
            f.write(content)

    @classmethod
    def deserialize(cls, content: bytes) -> SerializableType:
        for fmt in cls.FORMATS:
            magic = fmt.magic_string
            if content.startswith(magic):
                body = content[len(magic):]
                return pickle.loads(body)

        raise SerializationError("Unsupported format")

    @classmethod
    def load(cls, path: str) -> SerializableType:
        """
        Try to deserialize the content of the given file, based on its magic
        string.
        """
        with open(path, 'rb') as f:
            file_content = f.read()

        return cls.deserialize(file_content)

    @classmethod
    def magic(cls, object_type: type) -> bytes:
        for fmt in cls.FORMATS:
            if fmt.object_type == object_type:
                return fmt.magic_string

        msg = f"Serialization of {object_type} is not supported"
        raise SerializationError(msg)

    @classmethod
    def extension(cls, object_type: type) -> str:
        for fmt in cls.FORMATS:
            if fmt.object_type == object_type:
                return fmt.extension

        msg = f"Serialization of {object_type} is not supported"
        raise SerializationError(msg)

    @classmethod
    def type(cls, magic: bytes) -> "Serializer.SerializableType":
        for fmt in cls.FORMATS:
            if fmt.magic_string == magic:
                return fmt.object_type

        msg = f"Magic string {magic.decode('utf-8')} is not supported"
        raise SerializationError(msg)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['emu']

package_data = \
{'': ['*']}

install_requires = \
['oletools>=0.55.1,<0.56.0',
 'prettytable>=0.7.2,<0.8.0',
 'pyparsing>=2.4.7,<3.0.0',
 'python-magic>=0.4.18,<0.5.0']

entry_points = \
{'console_scripts': ['emu = emu.__main__:main']}

setup_kwargs = {
    'name': 'spurious-emu',
    'version': '0.2.0',
    'description': 'VBA static and dynamic analysis tool for malware analysts',
    'long_description': '# SpuriousEmu\n\nVisual Basic for Applications tools allowing to parse VBA files, interpret them and extract behaviour information for malware analysis purpose.\n\n## Installation\n\nSpuriousEmu is available on PyPI, so you can install it using\n\n```bash\npip install spurious-emu\n```\n\n## Usage\n\nSpuriousEmu can work with VBA source files, or directly with Office documents. For the later case, it relies on olevba to extract macros from the files. All of the command use a final positional argument to specify the input file to work with.\n\nIf you work with VBA source files, the following convention is used:\n    - procedural modules have `.bas` extension\n    - class modules have `.cls` extension\n    - standalone script files have `.vbs` extension\n\nSpuriousEmu uses different subcommands for its different operating modes.\n\n### Static analysis\n\nStatic analysis is performed using the `static` subcommand.\n\nUsually, the first step is to determine the different functions and classes defined, in order to understand the structure of the program. You can for example use it to determine the entry point prior to dynamic analysis. It is the default behaviour when using no flag:\n\n```bash\nemu static document.xlsm\n```\n\nAdditionally, for large files, you can use the `-o` flag to serialize the information compiled during static analysis into a binary file that you will be able to use later with the `report` command for example:\n\n```bash\nemu static -o document.spurious-com document.xlsm\n```\n\nYou can also de-obfuscate a file by using the `-d` flag, which specifies the de-obfuscation level. You can output the whole file, or a single function or module using the `-e` flag. The result can be sent to standard output or written to a file specified with the `-o` file:\n\n```bash\nemu static -d3 -e VBAEnv.Default.Main -o Main.bas document.xlsm\n```\n\n### Dynamic analysis\n\nYou can trigger dynamic analysis with the `dynamic` subcommand.\n\nOnce you have found the entry-point you want to use with the `static` subcommand, you can execute a file by specifying it with the `-e` flag. For example, to launch the `Main` function found in `doc.xlsm`, use\n\n```bash\nemu dynamic -e Main doc.xlsm\n```\n\nThis will display a report of the execution of the program. Additionally, if you want to save the files created during execution, you can use the `-o` flag: it specifies a directory to save files to. Each created file is then stored in a file with its md5 sum as file name, and a `{hash}.filename.txt` file contains its original name. You can also save a report of the dynamic analysis using the `-r` flag. For example:\n\n```bash\nemu dynamic -o extract_files -r report.spemu-out doc.xlsm\n```\n\n### Report production\n\nYou can work with `.spemu-out` and `.spemu-com` file with the `report` command.\n\nThe `report` commands can have three mutually exclusive flags: `--json`, `--csv` and `--table`, which change the way reports are displayed.\n\nSimilarly to the default `static` output, you can use the `--symbols` flag with a `.spemu-com` file to get the list of functions and classes. For example, to have them in a JSON dump, you can use\n\n```bash\nemu report --symbols --json program.spemu-com\n```\n\nYou can extract the files generated by the execution of a program using the `--extract-files` flag, which behaves like the `-o` flag with the `dynamic` command:\n\n```bash\nemu report --extract-files files program.spemu-out\n```\n\nA timeline of the events can be produced with the `--timeline` flag. It can be made easier to read with the `--shorten` and `--skip-streaks` commands, as in\n\n```bash\nemu report --timeline --table --shorten --skip-streaks 10 program.spemu-out\n```\n\n## Dependencies\n\nPython 3.8 is used, and SpuriousEmu mainly relies on [PyParsing](https://github.com/pyparsing/pyparsing) for VBA grammar parsing, and [oletools](http://www.decalage.info/python/oletools) to extract VBA macros from Office documents. Report tables are generated using [PrettyTable](https://github.com/jazzband/prettytable).\n\n[nose](https://nose.readthedocs.io/en/latest/man.html) is used as testing framework, and [mypy](http://mypy-lang.org/) to perform static code analysis. `lxml` and `coverage` are used to produce test reports.\n\n## Tests\n\nTo set a development environment up, use `poetry`:\n\n```bash\npoetry install\n```\n\nThen, use nose to run the test suite:\n\n```bash\npoetry run nosetests\n```\n\nAll test files are in `tests`, including:\n    - Python test scripts, starting with `test_`\n    - VBA scripts used to test the different stages of the tools, with `vbs` extensions, stored in `source`\n    - expected test results, stored as JSON dumps in `result`\n\n\nYou can use mypy to perform code static analysis:\n\n```bash\npoetry run mypy emu/*.py\n```\n\nBoth commands produce HTML reports stored in `tests/report`.\n',
    'author': 'Louis Dubois',
    'author_email': 'ldbo@protonmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ldbo/SpuriousEmu',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""Provide the Interpreter class, that is used to execute VBA programs."""

from typing import List, Optional

from .function import Function, InternalFunction, ExternalFunction
from .abstract_syntax_tree import *
from .compiler import Program
from .error import InterpretationError, ResolutionError
from .reference import Reference, Environment, Variable, FunctionReference
from .side_effect import Memory, OutsideWorld
from .operator import OPERATORS_MAP, Operator
from .value import Value, Integer, Object
from .visitor import Visitor


class Resolver(Visitor):
    """Used to tell the interpreter what each name referes to"""
    _interpreter: "Interpreter"
    _program: Program
    _current_reference: Reference
    _previous_references: List[Reference]
    _resolution: Optional[Reference]

    def __init__(self, interpreter, program=None) -> None:
        self._interpreter = interpreter
        self._program = program if program is not None \
            else Program(Memory(), Environment())
        self._current_reference = self._program.environment
        self._previous_references = []
        self._resolution = None

    def visit_Identifier(self, identifier: Identifier) -> None:
        self._resolution = Resolver.resolve_from(
            self._current_reference, identifier.name)

    def visit_Get(self, get: Get) -> None:
        """Resolve a composite symbol."""
        parent_resolution = self.resolve(get.parent)
        self._resolution = parent_resolution.get_child(get.child.name)

    def resolve(self, symbol: Union[Identifier, Get]) -> Reference:
        """
        Translate a given simple identifier into the corresponding reference.

        :raises ResolutionError: If the resolution can't be done
        """
        self.visit(symbol)
        resolution = self._resolution
        self._resolution = None
        return resolution

    @staticmethod
    def resolve_from(reference: Reference, name: str, go_down: bool = False,
                     exclude: Reference = None) -> Reference:
        # TODO efficient search
        try:
            return reference.search_child(name)
        except ResolutionError:
            pass

        if reference.name == name:
            return reference

        if reference.parent is None:
            msg = f"Symbol {name} is not found in {reference}"
            raise ResolutionError(msg)

        return Resolver.resolve_from(reference.parent, name,
                                     go_down=True, exclude=reference)

    def find_function(self, function_name) -> List[Function]:
        functions = []
        for project in self._program.environment.children:
            for module in project.children:
                try:
                    functions.append(module.get_function(function_name))
                except ResolutionError:
                    pass

        return functions

    def jump(self, reference) -> None:
        """Change  to a new scope"""
        self._previous_references.append(self._current_reference)
        self._current_reference = reference

    def jump_back(self) -> None:
        """Go back to the last scope"""
        self._current_reference = self._previous_references.pop()


class Interpreter(Visitor):
    """Class used to interprete a program which has already been compiled."""
    _memory: Memory
    _resolver: Resolver
    _outside_world: OutsideWorld
    _evaluation: Optional[Value]

    # Main API
    def __init__(self, program: Optional[Program] = None):
        """
        Create an interpreter used to run a precompiled program, or simply to
        run an AST starting without a state.
        """
        if program is None:
            program = Program(Memory(), Environment())

        self._memory = program.memory
        self._resolver = Resolver(self, program)
        self._outside_world = OutsideWorld()
        self._evaluation = None

    def run(self, function_name: str, args: Optional[List[Value]] = None) \
            -> None:
        """
        Search for a function entry point in the program references, and
        execute it. You should use it most of the time.
        """
        if args is None:
            args = []

        functions = self._resolver.find_function(function_name)
        for function in functions:
            self._resolver._current_reference = function
            self.call_function(self._memory.get_function(str(function)), args)

    def print_stdout(self, state: bool) -> None:
        """Enable or disable stdout forwarding"""
        if state:
            hook = print
        else:
            def hook(*args, **kwargs):
                pass

        self._outside_world.add_hook(OutsideWorld.Event.Category.STDOUT, hook)

    # Internal interface

    def evaluate(self, expression: Expression) -> Value:
        """Return the value of an expression"""
        assert(isinstance(expression, Expression))

        self.visit(expression)
        value = self._evaluation
        self._evaluation = None
        return value

    def visit_Block(self, block: Block) -> None:
        for statement in block.body:
            try:
                self.visit(statement)
            except InterpretationError as e:
                raise e
            except AssertionError as e:
                raise e
            except Exception as e:
                msg = f"{statement.file}:{statement.line_number}: {e}"
                raise InterpretationError(msg)

    def visit_VarDec(self, var_dec: VarDec) -> None:
        # TODO
        # Add initial value
        pass

    def visit_VarAssign(self, var_assign: VarAssign) -> None:
        value = self.evaluate(var_assign.value)
        self._memory.set_variable(var_assign.variable.name, value)

    def visit_Literal(self, literal: Literal) -> None:
        self._evaluation = Value.from_literal(literal)

    def visit_Identifier(self, identifier: Identifier) -> None:
        resolution = self._resolver.resolve(identifier)
        if isinstance(resolution, Variable):
            self._evaluation = self._memory.get_variable(str(resolution))
        elif isinstance(resolution, FunctionReference):
            self._evaluation = self._memory.functions[str(resolution)]
        else:
            msg = f"Can't interprete identifier {identifier.name}"
            raise InterpretationError(msg)

    def visit_Get(self, get: Get) -> None:
        """
        Compute the value of a Get expression.
        """
        child_name = get.child.name

        # Get parent value
        if isinstance(get.parent, Get):
            try:
                parent = self._resolver.resolve(get.parent)
            except ResolutionError:
                parent = self.evaluate(get.parent)
        elif isinstance(get.parent, Identifier):
            parent = self._resolver.resolve(get.parent)
        elif isinstance(get.parent, FunCall):
            parent = self.evaluate(get.parent)
        else:
            msg = f"Get node with {type(parent)} parent is not supported"
            raise InterpretationError(msg)

        # Access child
        if isinstance(parent, Value):
            self.__access_member(parent, child_name)
        elif isinstance(parent, Reference):
            if isinstance(parent, Variable):
                parent_value = self._memory.get_variable(str(parent))
                self.__access_member(parent_value, child_name)
            elif parent.category is Reference.Category.Structural:

                child = parent.get_child(child_name)
                assert(isinstance(child, (Variable, FunctionReference)))
                if isinstance(child, Variable):
                    self._evaluation = self._memory.get_variable(str(child))
                elif isinstance(child, FunctionReference):
                    self._evaluation = self._memory.functions[str(child)]
        else:
            msg = f"Evaluation error, with parent {get.parent} and " \
                + f"child {get.child}"
            raise InterpretationError(msg)

    def __access_member(self, object_value: Value, child_name: str) \
            -> None:
        """
        Helper function to access a variable/method from an object, putting its
        value in self._evaluation.
        """
        assert(isinstance(object_value, Object))
        try:
            variable = object_value.variables[child_name]
            self._evaluation = variable
        except KeyError:
            class_ref = object_value.class_reference
            method_reference = class_ref.get_child(child_name)
            unbound_method = self._memory.functions[str(method_reference)]
            bound_method = unbound_method.create_bound_method(object_value)
            self._evaluation = bound_method

    def visit_UnOp(self, un_op: UnOp) -> None:
        arg_value = self.evaluate(un_op.argument)
        op = Operator.from_symbol(un_op.operator)
        self._evaluation = op.operate(arg_value)

    def visit_BinOp(self, bin_op: BinOp) -> None:
        left_value = self.evaluate(bin_op.left)
        right_value = self.evaluate(bin_op.right)

        op = Operator.from_symbol(bin_op.operator)
        self._evaluation = op.operate(left_value, right_value)

    def visit_FunCall(self, fun_call: FunCall) -> None:
        # TODO handle method call
        # Resolution
        function = self.evaluate(fun_call.function)
        assert(isinstance(function, Function))

        # Arguments handling
        arg_list = fun_call.arguments.args
        arg_values = list(map(self.evaluate, arg_list))

        # Move to the function
        return_value = self.call_function(function, arg_values)

        self._evaluation = return_value

    def visit_If(self, conditional: If) -> None:
        done = False

        # If
        condition_value = self.evaluate(conditional.condition)
        if condition_value.convert_to(Type.Boolean).value:
            self.visit_Block(conditional)
            done = True

        # Else ifs
        if not done:
            for elseif in conditional.elsifs:
                condition_value = elseif.evaluate(elseif.condition)
                if condition_value.convert_to(Type.Boolean).value:
                    self.visit_Block(elseif)
                    done = True

        # Else
        done = done or conditional.else_block is None
        if not done:
            self.visit_Block(conditional.else_block)

    def visit_For(self, loop: For) -> None:
        # Evaluation of parameters
        # TODO check if Let-coercible to Double, raise E13 if not
        start_value = self.evaluate(loop.start)
        end_value = self.evaluate(loop.end)
        if loop.step is None:
            step_value = Integer(1)
        else:
            step_value = self.evaluate(loop.step)
        step_literal = Literal.from_value(step_value)
        counter_name = loop.counter.name

        # Comparison function
        if step_value.value >= 0:
            def keep_going():
                counter_value = self._memory.get_variable(counter_name)
                return counter_value.value <= end_value.value
        else:
            def keep_going():
                counter_value = self._memory.get_variable(counter_name)
                return counter_value.value > end_value.value

        # Loop
        self._memory.set_variable(counter_name, start_value)
        while keep_going():
            self.visit_Block(loop)
            add_expression = BinOp('+', loop.counter, step_literal)
            new_counter_value = self.evaluate(add_expression)
            self._memory.set_variable(counter_name, new_counter_value)

    def visit_OnError(self, on_error: OnError) -> None:
        # TODO
        pass

    def visit_Resume(self, resume: Resume) -> None:
        # TODO
        pass

    def visit_ErrorStatement(self, error: ErrorStatement) -> None:
        # TODO
        pass

    def call_function(self, function: Function,
                      arguments_values: List[Value]) -> Optional[Value]:
        # TODO handle attribute access for internal class method
        if type(function) is ExternalFunction:
            if function.parent_object is not None:
                arguments_values = [function.parent_object] + arguments_values
            python_value = function.external_function(self, arguments_values)
            return Value.from_python_base_type(python_value)
        elif type(function) is InternalFunction:
            # Init memory and load arguments
            self._resolver.jump(function.reference)
            self._memory.new_locals()
            for name, value in zip(function.arguments_names, arguments_values):
                self._memory.set_variable(name, value)

            # Execute function
            self.visit_Block(function)

            # Get return value
            return_value: Optional[Value]
            try:
                return_value = self._memory.get_variable(function.name)
            except KeyError:
                return_value = None

            # Clean memory
            self._memory.discard_locals()
            self._resolver.jump_back()

            return return_value
        else:
            msg = f"{type(function)} is not handled yet by the interpreter"
            raise NotImplemented(msg)

    def create_object(self, class_name: str) -> Object:
        """
        Instanciate a new object with None values for each of its
        attributes.

        :param class_name: Absolute name of the VBA class
        """
        vba_class = self._memory.classes[class_name]
        variables = {name: None for name in vba_class.variables}
        obj = Object(vba_class.class_reference, variables)
        return obj

    # External functions interface
    def add_stdout(self, content: str) -> None:
        self._outside_world.add_event(
            OutsideWorld.Event.Category.STDOUT,
            str(self._resolver._current_reference),
            content)

    def add_command_execution(self, command: str) -> None:
        self._outside_world.add_event(
            OutsideWorld.Event.Category.EXECUTION,
            str(self._resolver._current_reference),
            command
        )

    def add_file_event(self, event_type: str, path: str,
                       data: Optional[str] = None) -> None:
        event = {'type': event_type, 'path': path}
        if data is not None:
            event['data'] = data
        self._outside_world.add_event(
            OutsideWorld.Event.Category.FILESYSTEM,
            str(self._resolver._current_reference),
            event)

    def add_network_event(self, *args, **kwargs) -> None:
        # TODO
        pass

    @staticmethod
    def run_program(program: Program, entry_point: str = "Main") \
            -> OutsideWorld:
        """
        Execute the a program using the given entry_point.
        """
        interpreter = Interpreter(program)
        interpreter.run(entry_point)

        return interpreter._outside_world



"""SpuriousEmu interface"""

from .compiler import Compiler, Unit, Program
from .error import *
from .interpreter import Interpreter
from .preprocessor import Preprocessor
from .report import ReportGenerator
from .serialize import Serializer
from .syntax import Parser, AST
from .side_effect import OutsideWorld

__version__ = "0.2.0"



def CreateObject(interpreter, arguments):
    object_name = arguments[0].value

    if object_name == "WScript.Shell":
        return interpreter.create_object("VBAEnv.WSH.WshShell")
    elif object_name == "Scripting.FileSystemObject":
        return interpreter.create_object("VBAEnv.WSH.Filesystem")
    else:
        return None
